"""Module provider for Joker.com"""

# The Joker.com API is well documented.
# Several specificities compared to classic REST API in other providers:
#   - everything is done with GET requests: all parameters (including actual records data
#     and authentication data) are passed in the URLs,
#   - DNS zones are represented in a "strongly" formatted flat format (not JSON or XML) that may
#     remind a zone definition in bind9,
#   - every operation requires to pass the entire updated zone through the API, so care must be
#     taken to not alter unexpectedly other entries during the create/update/delete operations,
#   - all headers and data are contained in the response body; then this kind of body is composed
#     of several lines of type key: value containing the headers (including errors), then a blank
#     line makes the separation with the data itself (see _process_response for the body parsing).
import hashlib
import json
import logging
import re
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

API_BASE_URL = "https://dmapi.joker.com/request"
DNSZONE_ENTRY_PATTERN = re.compile(
    r"^(.+?)\s+(\w+)\s+(.+?)\s+((?:\\\"(?:.+?)\\\")|(?:.+?))\s+(\d+)(?:\s+(\d+)\s+(\d+)\s+(.+?)|\s+(.+?)|)$"
)


class _Response:
    def __init__(self, headers, data):
        self.headers = headers
        self.data = data

    def __str__(self):
        return json.dumps({"headers": self.headers, "data": self.data})


class Provider(BaseProvider):
    @staticmethod
    def get_nameservers() -> List[str]:
        return ["ns.joker.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.description = """
The Joker.com provider requires a valid token for authentication.
You can create one in the section 'Manage Joker.com API access keys' of 'My Profile' in your Joker.com account.
"""
        parser.add_argument(
            "--auth-token",
            help="specify the API Key to connect to the Joker.com API",
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self._session_id = None

    def authenticate(self):
        auth_token = self._get_provider_option("auth_token")

        response = requests.post(API_BASE_URL + "/login", data={"api-key": auth_token})

        result = _process_response(response)
        self._session_id = result.headers["Auth-Sid"]

        response = self._post(
            "/query-domain-list", data={"pattern": self.domain, "showstatus": 1}
        )

        if not response.data:
            raise AuthenticationError(
                f"Domain {self.domain} is not registered with this account."
            )

        data = response.data[0]
        items = data.split(" ")
        domain_status = items[2].split(",")

        if len(set(domain_status).difference(["production", "lock", "autorenew"])) > 0:
            raise AuthenticationError(
                f"Current status for domain {self.domain} is: {items[2]}"
            )

        self.domain_id = self.domain

    def cleanup(self) -> None:
        pass

    def _request(self, action="POST", url="/", data=None, query_params=None):
        if not data:
            data = {}

        data["auth-sid"] = self._session_id

        response = requests.post(API_BASE_URL + url, data=data)
        return _process_response(response)

    def list_records(self, rtype=None, name=None, content=None):
        response = self._post("/dns-zone-get", data={"domain": self.domain_id})
        zone_data = _extract_zonedata(response.data)

        zone_data = [entry for entry in zone_data if entry["type"]]

        if rtype:
            zone_data = [entry for entry in zone_data if entry["type"] == rtype]
        if name:
            zone_data = [
                entry
                for entry in zone_data
                if self._full_name(entry["label"]) == self._full_name(name)
            ]
        if content:
            zone_data = [
                entry for entry in zone_data if entry["target"] == content.strip()
            ]

        records = [
            {
                "type": entry["type"],
                "name": self._full_name(entry["label"]),
                "ttl": entry["ttl"],
                "content": entry["target"],
                "id": self._identifier(entry),
            }
            for entry in zone_data
        ]

        LOGGER.debug("list_records: %s", records)

        return records

    def create_record(self, rtype, name, content):
        if not rtype or not name or not content:
            raise Exception(
                "Error, rtype, name and content are mandatory to create a record."
            )

        response = self._post("/dns-zone-get", data={"domain": self.domain_id})
        zonedata = _extract_zonedata(response.data)

        new_entry = {
            "label": self._relative_name(name),
            "type": rtype,
            "pri": 0,
            "target": content,
            "ttl": self._get_lexicon_option("ttl"),
            "valid-from": None,
            "valid-to": None,
            "parameters": None,
        }

        if any(
            entry
            for entry in zonedata
            if self._identifier(new_entry) == self._identifier(entry)
        ):
            LOGGER.debug(
                "create_record (ignored, duplicate): %s", self._identifier(new_entry)
            )
            return True

        zonedata.append(new_entry)

        self._apply_zonedata(zonedata)

        return True

    def update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier and not (rtype and name):
            raise Exception(
                "Error, either identifier or rtype + name are mandatory to update a record."
            )

        response = self._post("/dns-zone-get", data={"domain": self.domain_id})
        zonedata = _extract_zonedata(response.data)

        selector_info = (
            f"identifier={identifier}" if identifier else f"type={rtype},name={name}"
        )

        if identifier:
            to_update = [
                entry for entry in zonedata if self._identifier(entry) == identifier
            ]
        else:
            to_update = [
                entry
                for entry in zonedata
                if entry["type"] == rtype
                and self._full_name(entry["label"]) == self._full_name(name)
            ]

        if not to_update:
            raise Exception(f"Error, could not find a record for {selector_info}.")

        if len(to_update) > 1:
            raise Exception(
                f"Error, found more than one record for {selector_info}. "
                "Please use an identifier to select one explicitly."
            )

        to_update[0].update(
            {
                "type": rtype if rtype else to_update[0]["type"],
                "label": self._relative_name(name) if name else to_update[0]["label"],
                "target": content if content else to_update[0]["content"],
                "ttl": self._get_lexicon_option("ttl"),
            }
        )

        self._apply_zonedata(zonedata)

        return True

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if not identifier and not rtype:
            raise Exception(
                "Error, either rtype or identifier are mandatory to delete a record."
            )

        response = self._post("/dns-zone-get", data={"domain": self.domain_id})
        zonedata = _extract_zonedata(response.data)

        if identifier:
            zonedata = [
                entry for entry in zonedata if self._identifier(entry) != identifier
            ]
        else:
            zonedata = [
                entry
                for entry in zonedata
                if entry["type"] is None
                or self._identifier(entry)
                != self._identifier(
                    {
                        "type": rtype,
                        "label": name if name else entry["label"],
                        "target": content if content else entry["target"],
                    }
                )
            ]

        self._apply_zonedata(zonedata)

        return True

    def _apply_zonedata(self, zonedata):
        data = []
        for entry in zonedata:
            if "raw" in entry:
                data.append(entry["raw"])
            else:
                # TXT entries always require content to be quoted
                target = (
                    f'"{entry["target"]}"'
                    if entry["type"] == "TXT"
                    else entry["target"]
                )
                line = f"{entry['label']} {entry['type']} {entry['pri']} {target} {entry['ttl']}"

                if entry["valid-from"] is not None and entry["valid-to"] is not None:
                    line = f"{line} {entry['valid-from']} {entry['valid-to']}"

                if entry["parameters"] is not None:
                    line = f"{line} {entry['parameters']}"

                data.append(line)

        self._post(
            "/dns-zone-put",
            data={"domain": self.domain_id, "zone": "\n".join(data)},
        )

    def _identifier(self, record):
        if "raw" in record:
            return None

        m = hashlib.sha256()
        m.update(("type=" + record.get("type", "") + ",").encode("utf-8"))
        m.update(
            ("name=" + self._full_name(record.get("label", "")) + ",").encode("utf-8")
        )
        m.update(("content=" + record.get("target", "") + ",").encode("utf-8"))

        return m.hexdigest()[0:7]

    def _relative_name(self, record_name):
        if record_name == self.domain_id:
            return "@"

        return super(Provider, self)._relative_name(record_name)

    def _full_name(self, record_name):
        if record_name == "@":
            return self.domain_id

        return super(Provider, self)._full_name(record_name)


def _extract_zonedata(data):
    processed = []

    for entry in data:
        match = DNSZONE_ENTRY_PATTERN.match(entry)

        if match:
            extracted = {
                "label": match.group(1),
                "type": match.group(2),
                "pri": match.group(3),
                "target": match.group(4),
                "ttl": int(match.group(5)),
            }

            try:
                extracted.update(
                    {
                        "parameters": match.group(9),
                        "valid-from": None,
                        "valid-to": None,
                    }
                )
            except IndexError:
                try:
                    extracted.update(
                        {
                            "parameters": match.group(8),
                            "valid-from": int(match.group(7)),
                            "valid-to": int(match.group(6)),
                        }
                    )
                except IndexError:
                    pass

            if extracted["type"] == "TXT":
                extracted["target"] = re.sub(
                    r'"(.*)"', r"\1", extracted["target"]
                ).strip()

            processed.append(extracted)
        else:
            processed.append(
                {
                    "raw": entry,
                    "type": None,
                }
            )

    return processed


def _process_response(response):
    response.raise_for_status()

    headers = {}
    body = []
    feed_headers = True

    data = response.text
    for line in data.split("\n"):
        if not line:
            feed_headers = False
            continue

        if feed_headers:
            items = line.split(":")
            headers[items[0]] = items[1].lstrip()
        else:
            body.append(line)

    processed_response = _Response(headers, body)

    if headers["Status-Code"] != "0":
        if body and body[0] == "API key is invalid":
            raise requests.exceptions.HTTPError(
                f"{headers['Status-Code']} Error: API key is invalid for url: {response.url}",
                response=processed_response,
            )
        raise requests.exceptions.HTTPError(
            f"{headers['Status-Code']} Error: {headers['Status-Text']} ({headers['Error']}) "
            f"for url: {response.url}",
            response=processed_response,
        )

    return processed_response



"""
Lexicon PowerDNS Provider

Author: Will Hughes, 2017

API Docs: https://doc.powerdns.com/md/httpapi/api_spec/

Implementation notes:
* The PowerDNS API does not assign a unique identifier to each record in the way
that Lexicon expects. We work around this by creating an ID based on the record
name, type and content, which when taken together are always unique
* The PowerDNS API has no notion of 'create a single record' or 'delete a single
record'. All operations are either 'replace the RRSet with this new set of records'
or 'delete all records for this name and type. Similarly, there is no notion of
'change the content of this record', because records are identified by their name,
type and content.
* The API is very picky about the format of values used when creating records:
** CNAMEs must be fully qualified
** TXT, LOC records must be quoted
This is why the _clean_content and _unclean_content methods exist, to convert
back and forth between the format PowerDNS expects, and the format Lexicon uses
"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class PowerDNSProviderError(Exception):
    """Generic PowerDNS exception"""


class Provider(BaseProvider):
    """Provider class for PowerDNS"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return []

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="specify token for authentication")
        parser.add_argument("--pdns-server", help="URI for PowerDNS server")
        parser.add_argument("--pdns-server-id", help="Server ID to interact with")
        parser.add_argument(
            "--pdns-disable-notify", help="Disable slave notifications from master"
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)

        self.api_endpoint = self._get_provider_option("pdns_server")
        if not self.api_endpoint:
            raise PowerDNSProviderError(
                "PowerDNS API endpoint not defined (pdns_server)"
            )

        self.disable_slave_notify = self._get_provider_option("pdns-disable-notify")

        if self.api_endpoint.endswith("/"):
            self.api_endpoint = self.api_endpoint[:-1]

        if not self.api_endpoint.endswith("/api/v1"):
            self.api_endpoint += "/api/v1"

        self.server_id = self._get_provider_option("pdns_server_id")
        if self.server_id is None:
            self.server_id = "localhost"

        self.api_endpoint += "/servers/" + self.server_id

        self.api_key = self._get_provider_option("auth_token")
        if not self.api_key:
            raise PowerDNSProviderError("PowerDNS API key not defined (auth_token)")

        self._zone_data = None

    def notify_slaves(self):
        """Checks to see if slaves should be notified, and notifies them if needed"""
        if self.disable_slave_notify is not None:
            LOGGER.debug("Slave notifications disabled")
            return False

        if self.zone_data()["kind"] == "Master":
            response_code = self._put("/zones/" + self.domain + "/notify").status_code
            if response_code == 200:
                LOGGER.debug("Slave(s) notified")
                return True
            LOGGER.debug("Slave notification failed with code %i", response_code)
        else:
            LOGGER.debug("Zone type should be 'Master' for slave notifications")
        return False

    def zone_data(self):
        """Get zone data"""
        if self._zone_data is None:
            self._zone_data = self._get(
                "/zones/" + self._ensure_dot(self.domain)
            ).json()
        return self._zone_data

    def authenticate(self):
        self.zone_data()
        self.domain_id = self.domain

    def cleanup(self) -> None:
        pass

    def _make_identifier(self, rtype, name, content):
        return f"{rtype}/{name}={content}"

    def _parse_identifier(self, identifier):
        parts = identifier.split("/")
        rtype = parts[0]
        parts = parts[1].split("=")
        name = parts[0]
        content = "=".join(parts[1:])
        return rtype, name, content

    def list_records(self, rtype=None, name=None, content=None):
        records = []
        for rrset in self.zone_data()["rrsets"]:
            if (
                name is None or self._fqdn_name(rrset["name"]) == self._fqdn_name(name)
            ) and (rtype is None or rrset["type"] == rtype):
                for record in rrset["records"]:
                    if content is None or record["content"] == self._clean_content(
                        rtype, content
                    ):
                        records.append(
                            {
                                "type": rrset["type"],
                                "name": self._full_name(rrset["name"]),
                                "ttl": rrset["ttl"],
                                "content": self._unclean_content(
                                    rrset["type"], record["content"]
                                ),
                                "id": self._make_identifier(
                                    rrset["type"], rrset["name"], record["content"]
                                ),
                            }
                        )
        LOGGER.debug("list_records: %s", records)
        return records

    def _clean_content(self, rtype, content):
        if rtype in ("TXT", "LOC"):
            if content[0] != '"':
                content = '"' + content
            if content[-1] != '"':
                content += '"'
        elif rtype == "CNAME":
            content = self._fqdn_name(content)
        return content

    def _unclean_content(self, rtype, content):
        if rtype in ("TXT", "LOC"):
            content = content.strip('"')
        elif rtype == "CNAME":
            content = self._full_name(content)
        return content

    def create_record(self, rtype, name, content):
        rname = self._fqdn_name(name)
        newcontent = self._clean_content(rtype, content)

        updated_data = {
            "name": rname,
            "type": rtype,
            "records": [],
            "ttl": self._get_lexicon_option("ttl") or 600,
            "changetype": "REPLACE",
        }

        updated_data["records"].append({"content": newcontent, "disabled": False})

        for rrset in self.zone_data()["rrsets"]:
            if rrset["name"] == rname and rrset["type"] == rtype:
                updated_data["ttl"] = rrset["ttl"]

                for record in rrset["records"]:
                    if record["content"] != newcontent:
                        updated_data["records"].append(
                            {
                                "content": record["content"],
                                "disabled": record["disabled"],
                            }
                        )
                break

        request = {"rrsets": [updated_data]}
        LOGGER.debug("request: %s", request)

        self._patch("/zones/" + self._ensure_dot(self.domain), data=request)
        self.notify_slaves()
        self._zone_data = None
        return True

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier is not None:
            rtype, name, content = self._parse_identifier(identifier)

        LOGGER.debug("delete %s %s %s", rtype, name, content)
        if rtype is None or name is None:
            raise Exception("Must specify at least both rtype and name")

        for rrset in self.zone_data()["rrsets"]:
            if rrset["type"] == rtype and self._fqdn_name(
                rrset["name"]
            ) == self._fqdn_name(name):
                update_data = rrset

                if "comments" in update_data:
                    del update_data["comments"]

                if content is None:
                    update_data["records"] = []
                    update_data["changetype"] = "DELETE"
                else:
                    new_record_list = []
                    for record in update_data["records"]:
                        if (
                            self._clean_content(rrset["type"], content)
                            != record["content"]
                        ):
                            new_record_list.append(record)

                    update_data["records"] = new_record_list
                    update_data["changetype"] = "REPLACE"
                break

        request = {"rrsets": [update_data]}
        LOGGER.debug("request: %s", request)

        self._patch("/zones/" + self._ensure_dot(self.domain), data=request)
        self.notify_slaves()

        self._zone_data = None
        return True

    def update_record(self, identifier, rtype=None, name=None, content=None):
        self.delete_record(identifier)
        return self.create_record(rtype, name, content)

    def _patch(self, url="/", data=None, query_params=None):
        return self._request("PATCH", url, data=data, query_params=query_params)

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers={"X-API-Key": self.api_key, "Content-Type": "application/json"},
        )
        LOGGER.debug("response: %s", response.text)
        response.raise_for_status()
        return response

    @classmethod
    def _ensure_dot(cls, text):
        """
        This function makes sure a string contains a dot at the end
        """
        if text.endswith("."):
            return text
        return text + "."



"""Module provider for Yandex.Cloud DNS

API doc: https://cloud.yandex.com/en/docs/dns/

If you are using the service account, dns.viewer required for viewing the DNS records
and dns.editor for editing them.

If Cloud ID and Folder ID are not specified,
resource-manager.clouds.member and resource-manager.viewer are needed as well.
"""

import hashlib
import json
import logging
import urllib
from argparse import ArgumentParser
from typing import Dict, List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

__author__ = "Dmitry Verkhoturov"
__license__ = "MIT"
__contact__ = "https://github.com/paskal"

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for Yandex Cloud"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["yandexcloud.net"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-token",
            help="specify the IAM token (https://cloud.yandex.com/en/docs/dns/api-ref/authentication)",
        )
        parser.add_argument(
            "--dns-zone-id",
            help="specify the DNS Zone ID (can be obtained from web interface)",
        )
        parser.add_argument(
            "--cloud-id",
            help="specify the Cloud ID (visible in the cloud selector in the web interface), might be needed"
            " if DNS zone ID is not set",
        )
        parser.add_argument(
            "--folder-id",
            help="specify the Folder ID (https://cloud.yandex.com/en/docs/resource-manager/operations/folder/get-id) "
            "might be needed if DNS zone ID is not set",
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.default_ttl = 600  # 10 minutes: default value, used in Yandex.Cloud DNS
        self.api_endpoint = "https://dns.api.cloud.yandex.net/dns/v1"
        self.resources_api_endpoint = (
            "https://resource-manager.api.cloud.yandex.net/resource-manager/v1"
        )

    def _full_name_with_dot(self, s: str) -> str:
        return self._full_name(s) + "."

    @staticmethod
    def _identifier(rtype: str, name: str, content: str) -> str:
        sha256 = hashlib.sha256()
        sha256.update(("type=" + rtype + ",").encode("utf-8"))
        sha256.update(("name=" + name + ",").encode("utf-8"))
        sha256.update(("data=" + content).encode("utf-8"))
        return sha256.hexdigest()[0:7]

    def authenticate(self):
        dns_zone_id = self._get_provider_option("dns_zone_id")
        if dns_zone_id and self._cloud_id_matches_domain(dns_zone_id):
            self.domain_id = dns_zone_id

        # in case DNS zone ID is not set or set but not valid, fall back to retrieving it from the list
        if not self.domain_id:
            cloud_id = self._get_cloud_id()
            folder_id = self._get_folder_id(cloud_id)
            self.domain_id = self._get_dns_zone_id(folder_id)

    def cleanup(self) -> None:
        pass

    # verifies that the domain ID matches the domain in question
    def _cloud_id_matches_domain(self, dns_zone_id: str) -> bool:
        payload: Dict = self._get(f"{self.api_endpoint}/zones/{dns_zone_id}")
        if "zone" not in payload:
            # No DNS zone with given ID found
            return True
        if payload["zone"] != f"{self.domain}.":
            # Provided DNS zone ID does not match the domain
            return False
        return True

    def _get_cloud_id(self) -> str:
        """Gets Cloud ID from Resource Manager API
        https://cloud.yandex.com/en/docs/resource-manager/api-ref/Cloud/list"""
        cloud_id = self._get_provider_option("cloud_id")
        if cloud_id:
            return cloud_id

        payload: Dict = self._get(f"{self.resources_api_endpoint}/clouds")

        if "clouds" not in payload:
            raise AuthenticationError("No clouds found")
        if len(payload["clouds"]) > 1:
            raise AuthenticationError(
                "Too many clouds found, you must specify one explicitly, or set DNS zone ID"
            )

        return payload["clouds"][0]["id"]

    def _get_folder_id(self, cloud_id: str) -> str:
        """Gets Folder ID from Resource Manager API
        https://cloud.yandex.com/en/docs/resource-manager/api-ref/Folder/list
        """
        folder_id = self._get_provider_option("folder_id")
        if folder_id:
            return folder_id

        payload: Dict = self._get(
            f"{self.resources_api_endpoint}/folders", {"cloudId": cloud_id}
        )

        if "folders" not in payload:
            raise AuthenticationError(f"No folders found for Cloud ID {cloud_id}")
        if len(payload["folders"]) > 1:
            raise AuthenticationError(
                "Too many folders found, you must specify one explicitly, or set DNS zone ID"
            )

        return payload["folders"][0]["id"]

    def _get_dns_zone_id(self, folder_id: str) -> str:
        """Gets Domain ID from DnsZone API
        https://cloud.yandex.com/en/docs/dns/api-ref/DnsZone/list
        """
        payload: Dict = self._get(
            "/zones", {"folderId": folder_id, "filter": f"zone='{self.domain}.'"}
        )

        if "dnsZones" not in payload:
            raise AuthenticationError(f"No domain found for Folder ID {folder_id}")
        if len(payload["dnsZones"]) > 1:
            raise AuthenticationError("Too many domains found, this should not happen")

        return payload["dnsZones"][0]["id"]

    def list_records(self, rtype=None, name=None, content=None) -> List[Dict]:
        """List all records. Return an empty list if no records found.
        type, name and content are used to filter records.
        https://cloud.yandex.com/en/docs/dns/api-ref/DnsZone/listRecordSets
        """
        records: List[Dict] = []

        query_params = {"filter": ""}
        if rtype:
            query_params["filter"] = f"type='{rtype}'"

        # name should be in full form, transform it in case it's not
        if name:
            name = self._full_name_with_dot(name)
            query_params["filter"] += " AND " if query_params["filter"] else ""
            query_params["filter"] += f"name='{name}'"

        url = f"/zones/{self.domain_id}:listRecordSets"
        more_pages = True
        while more_pages:
            payload: Dict = self._get(url, query_params)
            if "nextPageToken" in payload:
                query_params["pageToken"] = payload["nextPageToken"]
            else:
                more_pages = False

            # nothing to list
            if "recordSets" not in payload:
                break

            for record in payload["recordSets"]:
                for entry in record["data"]:
                    # artificially generate ID for deletion and update
                    records.append(
                        {
                            "id": self._identifier(
                                record["type"], record["name"], entry
                            ),
                            "name": self._full_name(record["name"]),
                            "type": record["type"],
                            "ttl": record["ttl"],
                            "content": entry,
                        }
                    )

        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    def _get_record(self, identifier="", rtype="", name="") -> Dict:
        """Gets a single record by rtype and name, or by identifier.
        Throws HTTPError with status code 404 if record not found.
        https://cloud.yandex.com/en/docs/dns/api-ref/DnsZone/getRecordSet"""

        # find entry by identifier if it was provided.
        # the output format for _list_records and _get_record is different,
        # so it's necessary to find an entry in _list_records and then
        # re-request it in this function
        if identifier:
            for record in self.list_records():
                if record["id"] == identifier:
                    rtype = record["type"]
                    name = record["name"]

        if name is not None:
            name = self._full_name_with_dot(name)

        url = f"/zones/{self.domain_id}:getRecordSet"
        query_params = {"name": name, "type": rtype}
        # name should be in full form, transform it in case it's not

        return self._get(url, query_params)

    def create_record(self, rtype, name, content):
        """Create a record. Return True if record was created, or already the same precise record exists,
        False if record already exists or there is another problem.
        All params are required.
        """

        # name should be in full form, transform it in case it's not
        if name is not None:
            name = self._full_name_with_dot(name)

        # get existing content
        data = [content]
        try:
            record = self._get_record(rtype=rtype, name=name)
            data = record["data"]
            # add new content entry only if it's not yet present
            if content not in record["data"]:
                data.append(content)
        except requests.exceptions.HTTPError as error:
            if error.response.status_code != 404:
                raise error

        payload = self._post(
            f"/zones/{self.domain_id}:upsertRecordSets",
            {
                "replacements": [
                    {
                        "name": name,
                        "type": rtype,
                        "ttl": self._get_lexicon_option("ttl") or self.default_ttl,
                        "data": data,
                    }
                ]
            },
        )

        return self._check_request_success(payload, "create_record")

    def update_record(self, identifier=None, rtype=None, name=None, content=None):
        """Update existing record."""

        record = {
            "name": name,
            "type": rtype,
            "ttl": self._get_lexicon_option("ttl") or self.default_ttl,
            "data": content.split("\n"),
        }

        try:
            if identifier:
                record = self._get_record(identifier)
                if content and content in record["data"] and len(record["data"]) > 1:
                    # identifier is associated with particular content line
                    for i, entry in enumerate(record["data"]):
                        if identifier == self._identifier(
                            record["type"], record["name"], entry
                        ):
                            del record["data"][i]
                            record["data"].extend(content.split("\n"))
                            break
                # if we have the name and type alongside with id, use them instead of retrieved ones
                if name:
                    record["name"] = name
                if rtype:
                    record["type"] = rtype
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                LOGGER.info("update_record: identifier=%s does not exist", identifier)
                return True
            raise error

        payload = self._post(
            f"/zones/{self.domain_id}:upsertRecordSets", {"replacements": [record]}
        )
        return self._check_request_success(payload, "update_record")

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """Delete an existing record. Do nothing if record does not exist."""
        # name should be in full form, transform it in case it's not
        if name is not None:
            name = self._full_name_with_dot(name)

        try:
            record = self._get_record(identifier, rtype, name)
        except requests.exceptions.HTTPError as error:
            if error.response.status_code == 404:
                LOGGER.info(
                    "delete_record: id=%s type=%s name=%s does not exist",
                    identifier,
                    rtype,
                    name,
                )
                return True
            raise error

        # remove content entry if it's one of the many, and update record instead of deleting it
        action = "deletions"
        if content and content in record["data"] and len(record["data"]) > 1:
            record["data"].remove(content)
            action = "replacements"

        # deletion requires full match on also TTL and data, which would require us to find such entry first
        payload = self._post(
            f"/zones/{self.domain_id}:upsertRecordSets", {action: [record]}
        )

        return self._check_request_success(payload, "delete_record")

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None) -> Dict:
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        auth_token = self._get_provider_option("auth_token")

        if not auth_token:
            raise Exception("auth_token must be specified.")

        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"Bearer {auth_token}",
        }

        if not url.startswith("https://"):
            url = self.api_endpoint + url

        # urllib required here as requests by default encodes spaces as plus symbol (+) instead of %20
        response = requests.request(
            action,
            url,
            params=urllib.parse.urlencode(query_params, quote_via=urllib.parse.quote),
            data=json.dumps(data),
            headers=default_headers,
        )
        # if the request fails for any reason, throw an error.
        # response.raise_for_status() will throw and error, so the line below is just logging
        if "code" in response.json():
            LOGGER.warning(
                "error making request: %d, %s",
                response.json()["code"],
                response.json()["message"],
            )
        response.raise_for_status()
        return response.json()

    @staticmethod
    def _check_request_success(payload, title):
        if "code" in payload:
            LOGGER.debug("%s: code %d, %s", title, payload["code"], payload["message"])
            return False

        LOGGER.debug("%s: %s", title, payload)
        return True



"""Provide support to Lexicon for AWS Route 53 DNS changes."""

import hashlib
import logging
import re
from argparse import ArgumentParser
from typing import List

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

try:
    import boto3  # type: ignore
    import botocore  # type: ignore
except ImportError:
    pass

LOGGER = logging.getLogger(__name__)


class RecordSetPaginator:
    """Paginate through complete list of record sets."""

    def __init__(self, r53_client, hosted_zone_id, max_items=None):
        """Initialize paginator."""
        self.r53_client = r53_client
        self.hosted_zone_id = hosted_zone_id
        self.max_items = max_items

    def get_record_sets(self, **kwargs):
        """Retrieve a page from API."""
        return self.r53_client.list_resource_record_sets(**kwargs)

    def get_base_kwargs(self):
        """Get base kwargs for API call."""
        kwargs = {"HostedZoneId": self.hosted_zone_id}
        if self.max_items is not None:
            kwargs.update({"MaxItems": str(self.max_items)})
        return kwargs

    def all_record_sets(self):
        """Generator to loop through current record set.

        Call next page if it exists.
        """
        is_truncated = True
        start_record_name = None
        start_record_type = None
        kwargs = self.get_base_kwargs()
        while is_truncated:
            if start_record_name is not None:
                kwargs.update(
                    {
                        "StartRecordName": start_record_name,
                        "StartRecordType": start_record_type,
                    }
                )
            result = self.get_record_sets(**kwargs)
            for record_set in result.get("ResourceRecordSets", []):
                yield record_set

            is_truncated = result.get("IsTruncated", False)

            start_record_name = result.get("NextRecordName", None)
            start_record_type = result.get("NextRecordType", None)


class Provider(BaseProvider):
    """Provide AWS Route 53 implementation of Lexicon Provider interface."""

    @staticmethod
    def get_nameservers() -> List[re.Pattern]:
        return [re.compile(r"^awsdns-\d+\.\w+$")]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-access-key", help="specify ACCESS_KEY for authentication"
        )
        parser.add_argument(
            "--auth-access-secret", help="specify ACCESS_SECRET for authentication"
        )
        parser.add_argument(
            "--private-zone",
            help=(
                "indicates what kind of hosted zone to use. If true, use "
                "only private zones. If false, use only public zones"
            ),
        )

        # Allow bypassing the zone-id lookup for complex use cases like delegated subdomain
        parser.add_argument(
            "--zone-id", help="the AWS HostedZone ID to use; e.g. 'A1B2ZABCDEFGHI'"
        )

        # TODO: these are only required for testing, we should figure out
        # a way to remove them & update the integration tests
        # to dynamically populate the auth credentials that are required.
        parser.add_argument(
            "--auth-username",
            help="alternative way to specify the ACCESS_KEY for authentication",
        )
        parser.add_argument(
            "--auth-token",
            help="alternative way to specify the ACCESS_SECRET for authentication",
        )

    def __init__(self, config):
        """Initialize AWS Route 53 DNS provider."""
        super(Provider, self).__init__(config)
        # Allow setting domain_id from the command line
        self.domain_id = self._get_provider_option("zone_id") or None
        self.private_zone = self._get_provider_option("private_zone")
        # instantiate the client
        self.r53_client = boto3.client(
            "route53",
            aws_access_key_id=self._get_provider_option("auth_access_key")
            or self._get_provider_option("auth_username"),
            aws_secret_access_key=self._get_provider_option("auth_access_secret")
            or self._get_provider_option("auth_token"),
        )

    def filter_zone(self, data):
        """Check if a zone is private"""
        if self.private_zone is not None:
            if data["Config"]["PrivateZone"] != self.str2bool(self.private_zone):
                return False

        if data["Name"] != f"{self.domain}.":
            return False

        return True

    @staticmethod
    def str2bool(input_string):
        """Convert a string to boolean"""
        return input_string.lower() in ("true", "yes")

    def authenticate(self):
        """Authenticate the credentials early in the process"""

        # if this was set via the command-line argument, we don't need to look it up
        if self.domain_id is None:
            self._lookup_hosted_zone()
        # per https://github.com/AnalogJ/lexicon/pull/481#issuecomment-623203995
        # authenticate should do something even if we already know the ZoneID.
        # This will raise an exception if the credentials are wrong.
        # However, we do not want to do this if they have not specified the
        # ZoneID, since this may break existing user IAM roles.
        else:
            self.r53_client.get_hosted_zone(Id=self.domain_id)

    def cleanup(self) -> None:
        pass

    def _lookup_hosted_zone(self):
        """Determine the hosted zone id for the domain."""

        try:
            is_truncated = True
            next_dns_name = None
            next_hz_id = None
            hosted_zones = []

            while is_truncated:
                if not next_dns_name:
                    new_zones = self.r53_client.list_hosted_zones_by_name()
                else:
                    new_zones = self.r53_client.list_hosted_zones_by_name(
                        DNSName=next_dns_name, HostedZoneId=next_hz_id
                    )

                hosted_zones.extend(new_zones.get("HostedZones"))
                is_truncated = new_zones.get("IsTruncated")

                if is_truncated:
                    next_dns_name = new_zones.get("NextDNSName")
                    next_hz_id = new_zones.get("NextHostedZoneId")

            hosted_zone = next(hz for hz in hosted_zones if self.filter_zone(hz))
            self.domain_id = hosted_zone.get("Id")
        except StopIteration:
            raise AuthenticationError("No domain found")

    def _change_record_sets(self, action, rtype, name, content):
        ttl = self._get_lexicon_option("ttl")
        resource_records = []
        if isinstance(content, list):
            for i in content:
                value = f'"{i}"' if rtype in ["TXT", "SPF"] else i
                resource_records.append({"Value": value})
        else:
            value = f'"{content}"' if rtype in ["TXT", "SPF"] else content
            resource_records.append({"Value": value})
        try:
            self.r53_client.change_resource_record_sets(
                HostedZoneId=self.domain_id,
                ChangeBatch={
                    "Comment": f"{action} using lexicon Route 53 provider",
                    "Changes": [
                        {
                            "Action": action,
                            "ResourceRecordSet": {
                                "Name": self._fqdn_name(name),
                                "Type": rtype,
                                "TTL": ttl if ttl is not None else 300,
                                "ResourceRecords": resource_records,
                            },
                        }
                    ],
                },
            )
            return True
        except botocore.exceptions.ClientError as error:
            if "Duplicate Resource Record" in error.response["Error"]["Message"]:
                # Duplicate resource, that have been a noop. This is expected.
                return True
            LOGGER.error(str(error), exc_info=True)
            return False

    def create_record(self, rtype, name, content):
        """Create a record in the hosted zone."""
        existing_records = self._list_record_sets(rtype, name)
        if existing_records:
            existing_record = existing_records[0]
            if isinstance(existing_records[0]["content"], list):
                return self._change_record_sets(
                    "UPSERT",
                    existing_record["type"],
                    existing_record["name"],
                    existing_record["content"] + [content],
                )
            return self._change_record_sets(
                "UPSERT", rtype, name, [existing_record["content"]] + [content]
            )
        return self._change_record_sets("CREATE", rtype, name, content)

    def update_record(self, identifier=None, rtype=None, name=None, content=None):
        """Update a record from the hosted zone."""
        if identifier:
            records = [
                record
                for record in self.list_records()
                if identifier == _identifier(record)
            ]
            if not records:
                raise ValueError(f"No record found for identifier {identifier}")
            record = records[0]
            rtype = record["type"]
            name = record["name"]

        existing_records = self._list_record_sets(rtype, name)
        if not existing_records:
            raise ValueError("No matching record to update was found.")

        for existing_record in existing_records:
            if isinstance(existing_record["content"], list):
                # Multiple values in record.
                LOGGER.warning(
                    "Warning, multiple records found for given parameters, "
                    "only first entry will be updated: %s",
                    existing_record,
                )
                new_content = existing_record["content"].copy()
                new_content[0] = content
            else:
                new_content = content

            self._change_record_sets(
                "UPSERT", existing_record["type"], existing_record["name"], new_content
            )

        return True

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """Delete a record from the hosted zone."""
        if identifier:
            matching_records = [
                record
                for record in self.list_records()
                if identifier == _identifier(record)
            ]
            if not matching_records:
                raise ValueError(f"No record found for identifier {identifier}")
            rtype = matching_records[0]["type"]
            name = matching_records[0]["name"]
            content = matching_records[0]["content"]

        existing_records = self._list_record_sets(rtype, name, content)
        if not existing_records:
            raise ValueError("No record found for the provided type, name and content")

        for existing_record in existing_records:
            if isinstance(existing_record["content"], list) and content is not None:
                # multiple values in record, just remove one value and only if it actually exist
                if content in existing_record["content"]:
                    existing_record["content"].remove(content)
                    self._change_record_sets(
                        "UPSERT",
                        existing_record["type"],
                        existing_record["name"],
                        existing_record["content"],
                    )
            else:
                # if only one record exist, or if content is not specified, remove whole record
                self._change_record_sets(
                    "DELETE",
                    existing_record["type"],
                    existing_record["name"],
                    existing_record["content"],
                )

        return True

    def list_records(self, rtype=None, name=None, content=None):
        """List all records for the hosted zone."""
        records = self._list_record_sets(rtype, name, content)

        flatten_records = []
        for record in records:
            if isinstance(record["content"], list):
                for one_content in record["content"]:
                    flatten_record = record.copy()
                    flatten_record["content"] = one_content
                    flatten_record["id"] = _identifier(flatten_record)
                    flatten_records.append(flatten_record)
            else:
                record["id"] = _identifier(record)
                flatten_records.append(record)

        LOGGER.debug("list_records: %s", records)

        return flatten_records

    def _list_record_sets(self, rtype=None, name=None, content=None):
        records = []
        paginator = RecordSetPaginator(self.r53_client, self.domain_id)
        for record in paginator.all_record_sets():
            record_content = []
            if rtype is not None and record["Type"] != rtype:
                continue
            if name is not None and record["Name"] != self._fqdn_name(name):
                continue
            if record.get("AliasTarget", None) is not None:
                record_content = [record["AliasTarget"].get("DNSName", None)]
            if record.get("ResourceRecords", None) is not None:
                record_content = [
                    _format_content(record["Type"], value["Value"])
                    for value in record["ResourceRecords"]
                ]
            if content is not None and content not in record_content:
                continue

            LOGGER.debug("record: %s", record)
            records.append(
                {
                    "type": record["Type"],
                    "name": self._full_name(record["Name"]),
                    "ttl": record.get("TTL", None),
                    "content": (
                        record_content[0]
                        if len(record_content) == 1
                        else record_content
                    ),
                }
            )
        return records

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # Helper _request is not used in Route53 provider
        pass


def _format_content(rtype, content):
    return content[1:-1] if rtype in ["TXT", "SPF"] else content


def _identifier(record):
    sha256 = hashlib.sha256()
    sha256.update(("type=" + record.get("type", "") + ",").encode("utf-8"))
    sha256.update(("name=" + record.get("name", "") + ",").encode("utf-8"))
    sha256.update(("data=" + record.get("data", "") + ",").encode("utf-8"))
    return sha256.hexdigest()[0:7]



"""Module provider for Arvancloud"""

import json
import logging
from argparse import ArgumentParser
from typing import Any, List, Optional

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for Arvancloud"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["arvancloud.ir"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-token",
            help="specify key for authentication (API Key)",
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://napi.arvancloud.ir/cdn/4.0"

    def authenticate(self):
        payload = self._get(f"/domains/{self.domain}")

        if payload["data"]["domain"] == self.domain:
            self.domain_id = self.domain
        else:
            raise AuthenticationError("No matching domain found")

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing.
    def create_record(self, rtype, name, content):
        record = self._to_r1c_record(rtype, name, content)

        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")

        created = True

        try:
            self._post(f"/domains/{self.domain_id}/dns-records", record)
        except requests.exceptions.HTTPError as err:
            # HTTP 422 is expected when a record with the same type and content is exists.
            if (
                err.response.status_code == 422
                and isinstance(err.response.json()["errors"], dict)
                and err.response.json()["errors"].get("name")
                and err.response.json()["errors"]["name"][0]
                == "DNS Record Data is duplicate."
            ):
                created = True
            else:
                raise Exception("create_record: %s", err.response.json()["errors"])

        LOGGER.debug("create_record: %s", created)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        filter = {"per_page": 20}

        if rtype:
            filter["type"] = rtype.lower()
        if name:
            filter["search"] = self._relative_name(name)

        records = []

        while True:
            payload = self._get(f"/domains/{self.domain_id}/dns-records", filter)

            LOGGER.debug("payload: %s", payload)

            for record in payload["data"]:
                processed_record = {
                    "type": record["type"].upper(),
                    "name": self._full_name(record["name"]),
                    "ttl": record["ttl"],
                    "content": self._parse_r1c_response(
                        record["type"].upper(), record["value"]
                    ),
                    "id": record["id"],
                }
                records.append(processed_record)

            if content:
                records = [record for record in records if record["content"] == content]

            pages = payload["meta"]["total"]
            page = payload["meta"]["current_page"]

            if page >= pages:
                break

            filter["page"] = page + 1

        LOGGER.debug("list_records: %s", records)
        LOGGER.debug("Number of records retrieved: %d", len(records))
        return records

    # Update a record.
    def update_record(self, identifier=None, rtype=None, name=None, content=None):
        if not identifier:
            records = self.list_records(rtype, name)

            if len(records) == 1:
                identifier = records[0]["id"]
            elif len(records) < 1:
                raise Exception(
                    "No records found matching type and name - won't update"
                )
            else:
                raise Exception(
                    "Multiple records found matching type and name - won't update"
                )

        record = self._to_r1c_record(rtype, name, content)

        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")

        try:
            self._put(f"/domains/{self.domain_id}/dns-records/{identifier}", record)
        except requests.exceptions.HTTPError as err:
            raise Exception("update_record: %s", err.response.json()["errors"])

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []

        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        try:
            for record_id in delete_record_id:
                self._delete(f"/domains/{self.domain_id}/dns-records/{record_id}")
        except requests.exceptions.HTTPError as err:
            raise Exception("delete_record: %s", err.response.json()["errors"])

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"{self._get_provider_option('auth_token')}",
        }
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers=headers,
        )

        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()

    # Takes record input and puts it into a format the Arvancloud API supports
    def _to_r1c_record(self, rtype, name, content):
        if rtype in ["LOC", "SSHFP"]:
            raise NotImplementedError(f"The {rtype} isn't available for Arvancloud")
        output = {"type": rtype}

        if name is not None:
            output["name"] = self._relative_name(name)

        content_split = content.split(" ")

        if rtype in ["A", "AAAA"]:
            if len(content_split) == 1:
                output.update(
                    {
                        "value": [{"ip": content_split[0]}],
                    }
                )
            else:
                output.update(
                    {
                        "value": [
                            {"ip": ip, "port": int(port), "weight": int(weight)}
                            for ip, port, weight in content_split
                        ],
                    }
                )

        output.update(
            {
                "CNAME": (
                    lambda: {
                        "value": {
                            "host": content_split[0],
                            "host_header": (
                                content_split[1]
                                if len(content_split) == 2
                                else "source"
                            ),
                        }
                    }
                ),
                "ANAME": (
                    lambda: {
                        "value": {
                            "location": content_split[0],
                            "host_header": content_split[1],
                        }
                    }
                ),
                "MX": (
                    lambda: {
                        "value": {
                            "host": content_split[0],
                            "priority": content_split[1],
                        }
                    }
                ),
                "NS": (lambda: {"value": {"host": content}}),
                "PTR": (lambda: {"value": {"domain": content}}),
                "SRV": (
                    lambda: {
                        "value": {
                            "target": content_split[0],
                            "port": content_split[1],
                            "weight": content_split[2],
                            "priority": content_split[3],
                        }
                    }
                ),
                "TXT": (lambda: {"value": {"text": content}}),
                "SPF": (lambda: {"value": {"text": content}}),
                "DKIM": (lambda: {"value": {"text": content}}),
            }.get(rtype, lambda: {})()
        )

        return output

    # Takes record's value and puts it into a format the lexicon supports
    def _parse_r1c_response(self, rtype: str, input: Any) -> Optional[str]:
        if rtype == "CNAME" or rtype == "NS":
            return input["host"]

        if rtype == "TXT":
            return input["text"]

        if rtype == "A":
            return ", ".join(
                [
                    " ".join(
                        (str(value) if value is not None and value != "" else "")
                        for value in item.values()
                    )
                    for item in input
                ]
            )

        return None



"""Module provider for Regfish"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for Regfish"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["regfish-ns.net"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-api-key",
            help="specify API key for authentication",
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.regfish.de"

    def authenticate(self):
        payload = self._get(f"/dns/{self.domain}/rr")
        if not payload["success"]:
            raise AuthenticationError("No domain found")
        self.domain_id = 0

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, rtype, name, content):
        content = self._format_content(rtype, content)
        data = {
            "type": rtype,
            "name": self._fqdn_name(name),
            "data": content,
        }
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")
        if self._get_lexicon_option("priority"):
            if self._get_lexicon_option("priority").isnumeric():
                data["priority"] = int(self._get_lexicon_option("priority"))

        payload = {"success": True}
        try:
            payload = self._post("/dns/rr", data)
        except requests.exceptions.HTTPError as err:
            json_res = err.response.json()
            if (
                not json_res["code"] if "code" in json_res else 0 != 15003
            ):  # ResourceRecordAlreadyExists
                raise

        if payload["code"] if "code" in payload else 0 != 0:
            return False

        LOGGER.debug("create_record: %s", payload["success"])
        return payload["success"]

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        records = []
        payload = self._get(f"/dns/{self.domain}/rr")

        LOGGER.debug("payload: %s", payload)

        for record in payload["response"]:
            processed_record = {
                "type": record["type"],
                "name": self._full_name(record["name"]),
                "ttl": record["ttl"],
                "content": record["data"],
                "id": record["id"],
            }
            records.append(processed_record)

        # Apply filters on rtype, name, and content if they are provided
        if rtype or name or content:
            records = [
                record
                for record in records
                if (not rtype or record["type"] == rtype)
                and (not name or record["name"] == self._full_name(name))
                and (
                    not content
                    or record["content"] == self._format_content(rtype, content)
                )
            ]

        LOGGER.debug("list_records after filtering: %s", records)
        LOGGER.debug("Number of records retrieved after filtering: %d", len(records))
        return records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier is None:
            records = self.list_records(rtype, name)
            if len(records) == 1:
                identifier = records[0]["id"]
            elif len(records) < 1:
                raise Exception(
                    "No records found matching type and name - won't update"
                )
            else:
                raise Exception(
                    "Multiple records found matching type and name - won't update"
                )

        content = self._format_content(rtype, content)
        data = {}
        if rtype:
            data["type"] = rtype
        if name:
            data["name"] = self._fqdn_name(name)
        if content:
            data["data"] = content
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        payload = self._patch(f"/dns/rr/{identifier}", data)

        LOGGER.debug("update_record: %s", payload["success"])
        return payload["success"]

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        # Check if there are records to delete
        if not delete_record_id:
            # No records found to delete, raise an error or return False
            raise Exception(f"No matching records found to delete {identifier}")

        for record_id in delete_record_id:
            self._delete(f"/dns/rr/{record_id}")

        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        headers = {"Content-Type": "application/json", "Acccept": "application/json"}
        headers["x-api-key"] = self._get_provider_option("auth_api_key")
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data) if data else None,
            headers=headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()

    def _format_content(self, rtype, content):
        """
        Special case handling from some record types that Regfish needs
        formatted differently

        Returns new values for the content and data properties to be sent
        on the request
        """
        if rtype == "CNAME":
            content = content.rstrip(".") + "."

        return content



"""Provide support to Lexicon for DNS changes for Gransy sites subreg.cz, regtons.com and \
regnames.eu."""

import collections
import logging
from argparse import ArgumentParser
from builtins import staticmethod
from typing import List

try:
    import zeep
except BaseException:
    pass

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for Gransy"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["gransy.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.description = (
            "DNS manipulation provider for Gransy sites "
            + "subreg.cz, regtons.com and regnames.eu."
        )
        parser.add_argument(
            "--auth-username", help="specify username for authentication"
        )
        parser.add_argument(
            "--auth-password", help="specify password for authentication"
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.ssid = None

        client = zeep.Client("https://subreg.cz/wsdl")
        self.api = client.service

    # Authenticate against provider,
    # Make any requests required to get the domain's id for
    # this provider, so it can be used in subsequent calls.
    # Should throw an error if authentication fails for any reason,
    # of if the domain does not exist.
    def authenticate(self):
        """Logs-in the user and checks the domain name"""
        if not self._get_provider_option(
            "auth_username"
        ) or not self._get_provider_option("auth_password"):
            raise Exception(
                "No valid authentication data passed, expected: auth-username and auth-password"
            )
        response = self._request_login(
            self._get_provider_option("auth_username"),
            self._get_provider_option("auth_password"),
        )
        if "ssid" in response:
            self.ssid = response["ssid"]
            domains = self.domains_list()
            if any((domain["name"] == self.domain for domain in domains)):
                self.domain_id = self.domain
            else:
                raise AuthenticationError(f"Unknown domain {self.domain}")
        else:
            raise AuthenticationError("No SSID provided by server")

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing.
    def create_record(self, rtype, name, content):
        """Creates a new unique record"""
        found = self.list_records(rtype=rtype, name=name, content=content)
        if found:
            return True

        record = self._create_request_record(
            None,
            rtype,
            name,
            content,
            self._get_lexicon_option("ttl"),
            self._get_lexicon_option("priority"),
        )

        self._request_add_dns_record(record)
        return True

    # Update a record. Identifier must be specified.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        """Updates a record. Name changes are allowed, but the record identifier will change"""
        if identifier is not None:
            if name is not None:
                records = self._list_records_internal(identifier=identifier)
                if len(records) == 1 and records[0]["name"] != self._full_name(name):
                    # API does not allow us to update name directly
                    self._update_record_with_name(records[0], rtype, name, content)
                else:
                    self._update_record_with_id(identifier, rtype, content)
            else:
                self._update_record_with_id(identifier, rtype, content)
        else:
            guessed_record = self._guess_record(rtype, name)
            self._update_record_with_id(guessed_record["id"], rtype, content)
        return True

    def _update_record_with_id(self, identifier, rtype, content):
        """Updates existing record with no sub-domain name changes"""
        record = self._create_request_record(
            identifier,
            rtype,
            None,
            content,
            self._get_lexicon_option("ttl"),
            self._get_lexicon_option("priority"),
        )

        self._request_modify_dns_record(record)

    def _update_record_with_name(self, old_record, rtype, new_name, content):
        """Updates existing record and changes it's sub-domain name"""
        new_type = rtype if rtype else old_record["type"]

        new_ttl = self._get_lexicon_option("ttl")
        if new_ttl is None and "ttl" in old_record:
            new_ttl = old_record["ttl"]

        new_priority = self._get_lexicon_option("priority")
        if new_priority is None and "priority" in old_record:
            new_priority = old_record["priority"]

        new_content = content
        if new_content is None and "content" in old_record:
            new_content = old_record["content"]

        record = self._create_request_record(
            None, new_type, new_name, new_content, new_ttl, new_priority
        )

        # This will be a different domain name, so no name collision should
        # happen. First create a new entry and when it succeeds, delete the old
        # one.
        self._request_add_dns_record(record)
        self._request_delete_dns_record_by_id(old_record["id"])

    # Delete an existing record.
    # If record does not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """Deletes an existing record"""
        to_delete_ids = list()
        if identifier:
            to_delete_ids.append(identifier)
        else:
            for record in self.list_records(rtype=rtype, name=name, content=content):
                to_delete_ids.append(record["id"])

        for to_delete_id in to_delete_ids:
            self._request_delete_dns_record_by_id(to_delete_id)
        return True

    def domains_list(self):
        """Get list of registered domains"""
        response = self._request_domains_list()
        return response["domains"] if "domains" in response else list()

    def _create_request_record(self, identifier, rtype, name, content, ttl, priority):
        """Creates record for Subreg API calls"""
        record = collections.OrderedDict()

        # Mandatory content

        # Just for update - not for creation
        if identifier is not None:
            record["id"] = identifier

        record["type"] = rtype

        # Just for creation - not for update
        if name is not None:
            record["name"] = self._relative_name(name)

        # Optional content
        if content is not None:
            record["content"] = content
        if ttl is not None:
            record["ttl"] = ttl
        if priority is not None:
            record["prio"] = priority
        return record

    def _create_response_record(self, response):
        """Creates record for lexicon API calls"""
        record = dict()
        record["id"] = response["id"]
        record["type"] = response["type"]
        record["name"] = self._full_name(response["name"])
        if "content" in response:
            record["content"] = response["content"] or ""
        if "ttl" in response:
            record["ttl"] = response["ttl"]
        if "prio" in response:
            record["priority"] = response["prio"]
        return record

    def _full_name(self, record_name):
        """Returns full domain name of a sub-domain name"""
        # Handle None and empty strings
        if not record_name:
            return self.domain
        return super(Provider, self)._full_name(record_name)

    def _relative_name(self, record_name):
        """Returns sub-domain of a domain name"""
        # Handle None and empty strings as None
        if not record_name:
            return None
        subdomain = super(Provider, self)._relative_name(record_name)
        return subdomain if subdomain else None

    # List all records. Return an empty list if no records found
    # identifier, type, name and content are used to filter records.
    def list_records(self, rtype=None, name=None, content=None):
        return self._list_records_internal(rtype=rtype, name=name, content=content)

    def _list_records_internal(
        self, identifier=None, rtype=None, name=None, content=None
    ):
        """Lists all records by the specified criteria"""
        response = self._request_get_dns_zone()
        if "records" in response:
            # Interpret empty string as None because zeep does so too
            content_check = content if content != "" else None
            name_check = self._relative_name(name)

            # Stringize the identifier to prevent any rtype differences
            identifier_check = str(identifier) if identifier is not None else None

            filtered_records = [
                record
                for record in response["records"]
                if (identifier is None or str(record["id"]) == identifier_check)
                and (rtype is None or record["type"] == rtype)
                and (name is None or record["name"] == name_check)
                and (
                    content is None
                    or ("content" in record and record["content"] == content_check)
                )
            ]
            records = [
                self._create_response_record(filtered_record)
                for filtered_record in filtered_records
            ]
        else:
            records = []
        return records

    def _guess_record(self, rtype, name=None, content=None):
        """Tries to find existing unique record by type, name and content"""
        records = self._list_records_internal(
            identifier=None, rtype=rtype, name=name, content=content
        )
        if len(records) == 1:
            return records[0]
        if len(records) > 1:
            raise Exception(
                "Identifier was not provided and several existing "
                f"records match the request for {rtype}/{name}"
            )
        raise Exception(
            "Identifier was not provided and no existing records match "
            f"the request for {rtype}/{name}"
        )

    def _request_login(self, login, password):
        """Sends Login request"""
        return self._request_internal("Login", login=login, password=password)

    def _request_domains_list(self):
        """Sends Domains_List request"""
        return self._request_internal("Domains_List")

    def _request_get_dns_zone(self):
        """Sends Get_DNS_Zone request"""
        return self._request_internal("Get_DNS_Zone", domain=self.domain)

    def _request_add_dns_record(self, record):
        """Sends Add_DNS_Record request"""
        return self._request_internal(
            "Add_DNS_Record", domain=self.domain, record=record
        )

    def _request_modify_dns_record(self, record):
        """Sends Modify_DNS_Record request"""
        return self._request_internal(
            "Modify_DNS_Record", domain=self.domain, record=record
        )

    def _request_delete_dns_record_by_id(self, identifier):
        """Sends Delete_DNS_Record request"""
        return self._request_internal(
            "Delete_DNS_Record", domain=self.domain, record={"id": identifier}
        )

    def _request_internal(self, command, **kwargs):
        """Make request parse response"""
        args = dict(kwargs)
        if self.ssid:
            args["ssid"] = self.ssid
        method = getattr(self.api, command)
        response = method(**args)
        if response and "status" in response:
            if response["status"] == "error":
                self._raise_error(
                    message=response["error"]["errormsg"],
                    major=response["error"]["errorcode"]["major"],
                    minor=response["error"]["errorcode"]["minor"],
                )
            if response["status"] == "ok":
                return response["data"] if "data" in response else dict()
            raise Exception("Invalid status found in SOAP response")
        raise Exception("Invalid response")

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # Default helper _request is not used in Subreg provider
        pass

    @staticmethod
    def _raise_error(major, minor, message):
        raise GransyError(major, minor, message)


class GransyError(Exception):
    """Specific error for Gransy provider"""

    def __init__(self, major, minor, message):
        self.major = int(major)
        self.minor = int(minor)
        self.message = message
        super(GransyError, self).__init__()

    def __str__(self):
        return f"Major: {self.major} Minor: {self.minor} Message: {self.message}"



# coding: utf-8
# Copyright (c) 2016, Jason Kulatunga
# Copyright (c) 2021, Oracle and/or its affiliates.
"""
The Oracle Cloud Infrastructure (OCI) provider can create, list, update and
delete records in any public DNS zone hosted in a tenancy located in any
region within the OCI commercial (OC1) realm.

No authentication details are required if the OCI CLI installed and the DEFAULT
profile configured in the ~/.oci/config file has the appropriate permission for
the target DNS zone or the equivalent OCI_CLI environment variables are set.

Otherwise, you can either set the required LEXICON_OCI_AUTH_* environment
variables or pass the required --auth-* command-line parameters. If you do not
have a configuration file, you must provide user, tenancy, region, key and
fingerprint details.

Set the --auth-type parameter to 'instance_principal' to use instance principal
authentication when running Lexicon on an Oracle Cloud Infrastructure compute
instance. This method requires permission to be granted via IAM policy to a
dynamic group that includes the compute instance.

See https://docs.oracle.com/en-us/iaas/Content/DNS/Concepts/dnszonemanagement.htm
for in-depth documentation on managing DNS via the OCI console, SDK or API.
"""
import logging
import os
from argparse import ArgumentParser
from pathlib import Path
from typing import List

import requests

from lexicon.exceptions import AuthenticationError, LexiconError
from lexicon.interfaces import Provider as BaseProvider

# oci is an optional dependency of lexicon; do not throw an ImportError if
# the dependency is unmet.
try:
    from oci._vendor.requests.exceptions import ConnectTimeout  # type: ignore
    from oci.auth.signers import InstancePrincipalsSecurityTokenSigner  # type: ignore
    from oci.config import from_file  # type: ignore
    from oci.exceptions import ConfigFileNotFound  # type: ignore
    from oci.exceptions import InvalidConfig, ProfileNotFound
    from oci.signer import Signer  # type: ignore
except ImportError:
    pass

LOGGER = logging.getLogger(__name__)

CONFIG_VARS = {
    "user",
    "tenancy",
    "fingerprint",
    "key_file",
    "key_content",
    "pass_phrase",
}


class Provider(BaseProvider):
    """
    Provider class for Oracle Cloud Infrastructure DNS
    """

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["dns.oraclecloud.net"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.description = """
        Oracle Cloud Infrastructure (OCI) DNS provider
        """
        parser.add_argument(
            "--auth-config-file",
            help="The full path including filename to an OCI configuration file.",
        )
        parser.add_argument(
            "--auth-profile",
            help="The name of the profile to use (case-sensitive).",
        )
        parser.add_argument(
            "--auth-user",
            help="The OCID of the user calling the API.",
        )
        parser.add_argument(
            "--auth-tenancy",
            help="The OCID of your tenancy.",
        )
        parser.add_argument(
            "--auth-fingerprint",
            help="The fingerprint for the public key that was added to the calling user.",
        )
        parser.add_argument(
            "--auth-key-content",
            help="The full content of the calling user's private signing key in PEM format.",
        )
        parser.add_argument(
            "--auth-key-file",
            help="The full path including filename to the calling user's private signing key in PEM format.",
        )
        parser.add_argument(
            "--auth-pass-phrase",
            help="If the private key is encrypted, the pass phrase must be provided.",
        )
        parser.add_argument(
            "--auth-region",
            help="An OCI region identifier. Select the closest region for best performance.",
        )
        parser.add_argument(
            "--auth-type",
            help="Valid options are 'api_key' (default) or 'instance_principal'.",
        )

    def __init__(self, config):
        """Initialize OCI DNS client."""
        super(Provider, self).__init__(config)
        self.domain_id = None
        self._signer = None

        file_location = (
            self._get_provider_option("auth_config_file")
            if self._get_provider_option("auth_config_file")
            else str(Path(Path.home() / ".oci" / "config"))
        )
        profile_name = (
            self._get_provider_option("auth_profile")
            if self._get_provider_option("auth_profile")
            else "DEFAULT"
        )
        auth_type = (
            self._get_provider_option("auth_type")
            if self._get_provider_option("auth_type")
            else "api_key"
        )

        if auth_type == "instance_principal":
            try:
                self._signer = InstancePrincipalsSecurityTokenSigner()
                region = self._signer.region
            except ConnectTimeout:
                raise

        else:
            oci_config = {}
            try:
                oci_config = from_file(
                    file_location=file_location, profile_name=profile_name
                )
            except (ConfigFileNotFound, ProfileNotFound):
                for var in CONFIG_VARS:
                    if os.environ.get(f"OCI_CLI_{var.upper()}"):
                        oci_config[var] = os.environ.get(f"OCI_CLI_{var.upper()}")

                    if self._get_provider_option(f"auth_{var}"):
                        oci_config[var] = self._get_provider_option(f"auth_{var}")

            region = (
                self._get_provider_option("auth_region")
                if self._get_provider_option("auth_region")
                else "us-ashburn-1"
            )
            if hasattr(oci_config, "region") is False:
                oci_config["region"] = region

            try:
                self._signer = Signer.from_config(oci_config)
            except InvalidConfig:
                raise

        self.endpoint = f"https://dns.{region}.oraclecloud.com/20180115"
        LOGGER.debug(f"Activated OCI provider with endpoint: {self.endpoint}")

    def authenticate(self):
        try:
            zone = self._get(f"/zones/{self.domain}")
            self.domain_id = zone["id"]
            self.zone_name = zone["name"]
        except requests.exceptions.HTTPError as e:
            LOGGER.error(
                f"Error: invalid zone or permission denied accessing {self.domain}"
            )
            raise AuthenticationError(e)

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing
    def create_record(self, rtype, name, content):
        name = self._full_name(name)
        patchset = {
            "items": [
                {
                    "operation": "ADD",
                    "domain": name,
                    "rtype": rtype,
                    "rdata": content,
                    "ttl": (
                        self._get_lexicon_option("ttl")
                        if self._get_lexicon_option("ttl")
                        else None
                    ),
                }
            ]
        }

        try:
            self._patch(f"/zones/{self.zone_name}/records/{name}", patchset)
            LOGGER.debug(f"OCI: created new {rtype} record for {name}.")
        except requests.exceptions.HTTPError:
            raise LexiconError("OCI Error: record not created.")

        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        query_params = {"limit": 100, "rtype": rtype}
        name = self._full_name(name) if name else None

        if name:
            payload = self._get(f"/zones/{self.zone_name}/records/{name}", query_params)
        else:
            payload = self._get(f"/zones/{self.zone_name}/records", query_params)

        records = []
        for item in payload["items"]:
            rdata = item["rdata"].strip('"')
            name = self._full_name(item["domain"])
            if content and content != rdata:
                continue

            record = {
                "type": item["rtype"],
                "name": name,
                "ttl": item["ttl"],
                "content": rdata,
                "id": item["recordHash"],
            }
            records.append(record)

        LOGGER.debug(
            f"OCI: listing {len(records)} {rtype} records from {self.zone_name}."
        )
        return records

    # Update a record. Identifier must be specified.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        name = self._full_name(name) if name else None
        if identifier:
            records = [
                record
                for record in self.list_records(rtype=rtype)
                if identifier == record["id"]
            ]
        else:
            records = self.list_records(rtype=rtype, name=name)

        if not records:
            raise LexiconError(
                f"OCI Error: unable to find {rtype} record for {name} to update."
            )
        elif len(records) > 1:
            LOGGER.warning(
                f"Warning: multiple {rtype} records found for {name} containing {content}. Updating the first record returned.",
            )

        identifier = records[0]["id"]
        domain = self._full_name(records[0]["name"])
        ttl = (
            self._get_lexicon_option("ttl") if self._get_lexicon_option("ttl") else None
        )

        patchset = {
            "items": [
                {"operation": "REMOVE", "recordHash": identifier},
                {"operation": "ADD", "rtype": rtype, "rdata": content, "ttl": ttl},
            ]
        }

        try:
            self._patch(f"/zones/{self.zone_name}/records/{domain}", patchset)
            LOGGER.debug(f"OCI: updated {rtype} record [{identifier}].")
        except requests.exceptions.HTTPError:
            raise LexiconError(f"OCI Error updating {rtype} record for {domain}")

        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        name = self._full_name(name) if name else None
        if not identifier and not content:
            try:
                records = self.list_records(rtype, name)
                if len(records) > 0:
                    self._delete(f"/zones/{self.zone_name}/records/{name}/{rtype}")
                    LOGGER.debug(f"OCI: deleted {rtype} recordset for {name}.")
                else:
                    return True
            except requests.exceptions.HTTPError:
                raise LexiconError(f"OCI Error deleting {rtype} recordset for {name}")

        else:
            records = self.list_records(rtype=rtype, name=name, content=content)

            if identifier:
                records = [record for record in records if record["id"] == identifier]

            if len(records) == 0:
                return True

            patchset = {
                "items": [
                    {"operation": "REMOVE", "recordHash": record["id"]}
                    for record in records
                ]
            }

            try:
                self._patch(f"/zones/{self.zone_name}/records", patchset)
                LOGGER.debug(f"OCI: deleted {rtype} record(s) from {self.zone_name}.")
            except requests.exceptions.HTTPError:
                raise LexiconError("Error deleting record(s)")

        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if not data and action != "DELETE":
            data = {}

        if not query_params:
            query_params = {}

        if not url.startswith(self.endpoint):
            url = self.endpoint + url

        response = requests.request(
            action,
            url,
            params=query_params,
            auth=self._signer,
            json=data if data else None,
        )
        response.raise_for_status()

        if response.content:
            results = response.json()
        else:
            return None

        # The opc-next-page header indicates there is another page of results
        if "opc-next-page" in response.headers:
            query_params["page"] = response.headers["opc-next-page"]
            response = requests.request(
                action,
                url,
                params=query_params,
                auth=self.auth,
                json=data if data else None,
            )
            response.raise_for_status()
            results += response.json()

        return results



"""Compatibility layer for Linode4 (old name for Linode)"""

import logging

from lexicon._private.providers.linode import Provider as LinodeProvider

LOGGER = logging.getLogger(__name__)


# TODO: This provider is kept for retro-compatibility purpose
class Provider(LinodeProvider):
    """Provider for Linode V4"""

    def __init__(self, config):
        LOGGER.error(
            "Linode4 provider is deprecated and will be removed in a future version of Lexicon."
        )
        LOGGER.error("Please use Linode provider instead.")
        super().__init__(config)



"""Module provider for Henet"""

import logging
import re
from argparse import ArgumentParser
from typing import List

from bs4 import BeautifulSoup
from requests import Session

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """
    he.net provider
    """

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["he.net"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.description = """A provider for Hurricane Electric DNS.
        NOTE: THIS DOES NOT WORK WITH 2-FACTOR AUTHENTICATION.
              YOU MUST DISABLE IT IF YOU'D LIKE TO USE THIS PROVIDER.
        """
        parser.add_argument(
            "--auth-username", help="specify username for authentication"
        )
        parser.add_argument(
            "--auth-password", help="specify password for authentication"
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain = self.domain
        self.domain_id = None
        self.session = None

    def authenticate(self):
        # Create the session GET the login page to retrieve a session cookie
        self.session = Session()
        self.session.get("https://dns.he.net/")

        # Hit the login page with authentication info to login the session
        login_response = self.session.post(
            "https://dns.he.net",
            data={
                "email": self._get_provider_option("auth_username") or "",
                "pass": self._get_provider_option("auth_password") or "",
            },
        )

        # Parse in the HTML, if the div containing the error message is found, error
        html = BeautifulSoup(login_response.content, "html.parser")
        if html.find("div", {"id": "dns_err"}) is not None:
            LOGGER.warning("HE login failed, check HE_USER and HE_PASS")
            return False

        # Make an authenticated GET to the DNS management page
        zones_response = self.session.get("https://dns.he.net")

        html = BeautifulSoup(zones_response.content, "html.parser")
        zone_img = html.find("img", {"name": self.domain, "alt": "delete"})

        # If the tag couldn't be found, error, otherwise, return the value of the tag
        if zone_img is None:
            LOGGER.warning("Domain %s not found in account", self.domain)
            raise AuthenticationError(f"Domain {self.domain} not found in account")

        self.domain_id = zone_img["value"]
        LOGGER.debug("HENET domain ID: %s", self.domain_id)
        return True

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing
    def create_record(self, rtype, name, content):
        LOGGER.debug("Creating record for zone %s", name)
        # Pull a list of records and check for ours
        records = self.list_records(rtype=rtype, name=name, content=content)
        if len(records) >= 1:
            LOGGER.warning("Duplicate record %s %s %s, NOOP", rtype, name, content)
            return True
        data = {
            "account": "",
            "menu": "edit_zone",
            "Type": rtype,
            "hosted_dns_zoneid": self.domain_id,
            "hosted_dns_recordid": "",
            "hosted_dns_editzone": "1",
            "Priority": "",
            "Name": name,
            "Content": content,
            "TTL": "3600",
            "hosted_dns_editrecord": "Submit",
        }
        ttl = self._get_lexicon_option("ttl")
        if ttl:
            if ttl <= 0:
                data["TTL"] = "3600"
            else:
                data["TTL"] = str(ttl)
        prio = self._get_lexicon_option("priority")
        if prio:
            if prio <= 0:
                data["Priority"] = "10"
            else:
                data["Priority"] = str(prio)
        self.session.post("https://dns.he.net/index.cgi", data=data)
        # Pull a list of records and check for ours
        records = self.list_records(name=name)
        if len(records) >= 1:
            LOGGER.info("Successfully added record %s", name)
            return True
        LOGGER.info("Failed to add record %s", name)
        return False

    # List all records. Return an empty list if no records found.
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is
    # received.
    def list_records(self, rtype=None, name=None, content=None):
        return self._list_records_internal(rtype=rtype, name=name, content=content)

    def _list_records_internal(
        self, rtype=None, name=None, content=None, identifier=None
    ):
        records = []
        # Make an authenticated GET to the DNS management page
        edit_response = self.session.get(
            f"https://dns.he.net/?hosted_dns_zoneid={self.domain_id}&menu=edit_zone&hosted_dns_editzone",
        )

        # Parse the HTML response, and list the table rows for DNS records
        html = BeautifulSoup(edit_response.content, "html.parser")

        def is_dns_tr_type(klass):
            return klass and re.compile("dns_tr").search(klass)

        records = html.find_all("tr", class_=is_dns_tr_type)

        # If the tag couldn't be found, error, otherwise, return the value of the tag
        if records is None or not records:
            LOGGER.warning("Domains not found in account")
            return records

        new_records = []
        for dns_tr in records:
            tds = dns_tr.find_all("td")
            # Process HTML in the TR children to derive each object
            rec = {}
            rec["zone_id"] = tds[0].string
            rec["id"] = tds[1].string
            rec["name"] = tds[2].string
            # the 4th entry is a comment
            type_elem = tds[3].find("span", class_="rrlabel")
            rec["type"] = type_elem.string if type_elem else None
            rec["ttl"] = tds[4].string
            if tds[5].string != "-":
                rec["priority"] = tds[5]
            rec["content"] = tds[6].string
            rec["is_dynamic"] = tds[7].string == "1"
            rec = self._clean_TXT_record(rec)
            new_records.append(rec)
        records = new_records
        if identifier:
            LOGGER.debug("Filtering %d records by id: %s", len(records), identifier)
            records = [record for record in records if record["id"] == identifier]
        if rtype:
            LOGGER.debug("Filtering %d records by rtype: %s", len(records), rtype)
            records = [record for record in records if record["type"] == rtype]
        if name:
            LOGGER.debug("Filtering %d records by name: %s", len(records), name)
            if name.endswith("."):
                name = name[:-1]
            records = [record for record in records if name in record["name"]]
        if content:
            LOGGER.debug(
                "Filtering %d records by content: %s", len(records), content.lower()
            )
            records = [
                record
                for record in records
                if record["content"].lower() == content.lower()
            ]
        LOGGER.debug("Final records (%d): %s", len(records), records)

        return records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        # Delete record if it exists
        self.delete_record(identifier, rtype, name, content)
        return self.create_record(rtype, name, content)

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_ids = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_ids = [record["id"] for record in records]
        else:
            delete_record_ids.append(identifier)
        LOGGER.debug("Record IDs to delete: %s", delete_record_ids)
        for rec_id in delete_record_ids:
            # POST to the DNS management UI with form values to delete the record
            delete_response = self.session.post(
                "https://dns.he.net/index.cgi",
                data={
                    "menu": "edit_zone",
                    "hosted_dns_zoneid": self.domain_id,
                    "hosted_dns_recordid": rec_id,
                    "hosted_dns_editzone": "1",
                    "hosted_dns_delrecord": "1",
                    "hosted_dns_delconfirm": "delete",
                },
            )

            # Parse the HTML response, if the <div> tag indicating success isn't found, error
            html = BeautifulSoup(delete_response.content, "html.parser")
            if html.find("div", {"id": "dns_status"}) is None:
                LOGGER.warning("Unable to delete record %s", rec_id)
                return False
        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # Helper _request is not used in this provider
        pass



"""Module provider for Sakura Cloud"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests
from requests.auth import HTTPBasicAuth

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for Sakura Cloud"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["sakura.ne.jp"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-token", help="specify access token for authentication"
        )
        parser.add_argument(
            "--auth-secret", help="specify access secret for authentication"
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://secure.sakura.ad.jp/cloud/zone/is1a/api/cloud/1.1"

    def authenticate(self):
        query_params = {"Filter": {"Provider.Class": "dns", "Name": self.domain}}
        payload = self._get("/commonserviceitem", query_params=query_params)

        for item in payload["CommonServiceItems"]:
            if item["Status"]["Zone"] == self.domain:
                self.domain_id = item["ID"]
                return

        raise AuthenticationError("No domain found")

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, rtype, name, content):
        name = self._relative_name(name)
        resource_record_sets = self._get_resource_record_sets()
        index = self._find_resource_record_set(
            resource_record_sets, rtype=rtype, name=name, content=content
        )
        if index >= 0:
            LOGGER.debug("create_record: %s", False)
            return False

        resource_record_sets.append(
            {
                "Name": name,
                "Type": rtype,
                "RData": self._bind_format_target(rtype, content),
                "TTL": self._get_lexicon_option("ttl"),
            }
        )

        self._update_resource_record_sets(resource_record_sets)
        LOGGER.debug("create_record: %s", True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        records = []

        for record in self._get_resource_record_sets():
            processed_record = {
                "type": record["Type"],
                "name": self._full_name(record["Name"]),
                "ttl": record["TTL"],
                "content": record["RData"],
                # 'id': None,
            }
            records.append(processed_record)

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def update_record(self, identifier=None, rtype=None, name=None, content=None):
        if not (rtype and name and content):
            raise Exception("rtype ,name and content must be specified.")

        name = self._relative_name(name)
        resource_record_sets = self._get_resource_record_sets()
        index = self._find_resource_record_set(
            resource_record_sets, rtype=rtype, name=name
        )

        if index >= 0:
            resource_record_sets[index]["Type"] = rtype
            resource_record_sets[index]["Name"] = name
            resource_record_sets[index]["RData"] = self._bind_format_target(
                rtype, content
            )
            resource_record_sets[index]["TTL"] = self._get_lexicon_option("ttl")
        else:
            resource_record_sets.append(
                {
                    "Name": name,
                    "Type": rtype,
                    "RData": self._bind_format_target(rtype, content),
                    "TTL": self._get_lexicon_option("ttl"),
                }
            )

        self._update_resource_record_sets(resource_record_sets)
        LOGGER.debug("create_record")

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        resource_record_sets = self._get_resource_record_sets()

        if name is not None:
            name = self._relative_name(name)
        if content is not None:
            content = self._bind_format_target(rtype, content)

        filtered_records = []
        for record in resource_record_sets:
            if rtype and record["Type"] != rtype:
                continue
            if name and record["Name"] != name:
                continue
            if content and record["RData"] != content:
                continue
            filtered_records.append(record)

        if not filtered_records:
            LOGGER.debug("delete_record: %s", False)
            return False

        for record in filtered_records:
            resource_record_sets.remove(record)

        self._update_resource_record_sets(resource_record_sets)
        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _full_name(self, record_name):
        if record_name == "@":
            record_name = self.domain
        return super(Provider, self)._full_name(record_name)

    def _relative_name(self, record_name):
        name = super(Provider, self)._relative_name(record_name)
        if not name:
            name = "@"
        return name

    def _bind_format_target(self, rtype, target):
        if rtype == "CNAME" and not target.endswith("."):
            target += "."
        return target

    def _find_resource_record_set(self, records, rtype=None, name=None, content=None):
        for index, record in enumerate(records):
            if rtype and record["Type"] != rtype:
                continue
            if name and record["Name"] != name:
                continue
            if content and record["RData"] != content:
                continue
            return index
        return -1

    def _get_resource_record_sets(self):
        payload = self._get(f"/commonserviceitem/{self.domain_id}")
        return payload["CommonServiceItem"]["Settings"]["DNS"]["ResourceRecordSets"]

    def _update_resource_record_sets(self, resource_record_sets):
        content = {
            "CommonServiceItem": {
                "Settings": {"DNS": {"ResourceRecordSets": resource_record_sets}}
            }
        }
        return self._put(f"/commonserviceitem/{self.domain_id}", content)

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        default_auth = HTTPBasicAuth(
            self._get_provider_option("auth_token"),
            self._get_provider_option("auth_secret"),
        )

        query_string = ""
        if query_params:
            query_string = json.dumps(query_params)

        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_string,
            data=json.dumps(data),
            headers=default_headers,
            auth=default_auth,
        )
        try:
            # if the request fails for any reason, throw an error.
            response.raise_for_status()
        except BaseException:
            LOGGER.error(response.json().get("error_msg"))
            raise
        return response.json()



"""Module provider for Flexible Engine Cloud"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for Flexible Engine Cloud"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["orange-business.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="specify token for authentication")
        parser.add_argument(
            "--zone-id",
            help="specify the zone id",
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.api_endpoint = "https://dns.prod-cloud-ocb.orange-business.com/v2"
        self.domain_id = None

    def authenticate(self):
        zone_id = self._get_provider_option("zone_id")
        payload = self._get("/zones", {"name": self.domain})

        if not zone_id:
            if not payload["zones"]:
                raise AuthenticationError("No domain found")
            if len(payload["zones"]) > 1:
                raise AuthenticationError(
                    "Too many domains found. This should not happen"
                )
            self.domain_id = payload["zones"][0]["id"]
            self.domain = payload["zones"][0]["name"].rstrip(".")
        else:
            self.domain_id = zone_id
            self.domain = payload["zones"][0]["name"].rstrip(".")

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype, name, content):
        # put string in array
        tmp = content
        content = []
        content.append(tmp)

        record = {"type": rtype, "name": self._full_name(name), "records": content}

        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")

        if rtype == "TXT":
            # Convert "String" to "\"STRING\""
            tmp = []
            tmp.append('"' + record["records"][0] + '"')
            record["records"] = tmp
        try:
            self._post(f"/zones/{self.domain_id}/recordsets", record)
        except requests.exceptions.HTTPError as err:
            already_exists = next(
                (
                    True
                    for error in err.response.json()
                    if err.response.json()["code"] == "DNS.0312"
                ),
                False,
            )
            if not already_exists:
                raise

        LOGGER.debug("create_record: %s", True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.

    def list_records(self, rtype=None, name=None, content=None):
        url = f"/zones/{self.domain_id}/recordsets"
        records = []
        payload = {}

        # Convert it to Array if it is not converted yet.
        if isinstance(content, str):
            tmp = content
            content = []
            content.append(tmp)

        # Iterating recordsets
        next_url = url
        while next_url is not None:
            payload = self._get(next_url)
            if "links" in payload and "next" in payload["links"]:
                next_url = payload["links"]["next"]
            else:
                next_url = None

            for record in payload["recordsets"]:
                processed_record = {
                    "type": record["type"],
                    "name": f"{record['name']}",
                    "ttl": record["ttl"],
                    "content": record["records"],
                    "id": record["id"],
                }
                records.append(processed_record)

        if rtype:
            records = [record for record in records if record["type"] == rtype]

        if name:
            records = [
                record
                for record in records
                if record["name"].rstrip(".") == self._full_name(name)
            ]

        if content:
            if len(content) > 1:
                records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    # update a record.

    def update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier is None:
            records = self.list_records(rtype, name)
            if len(records) == 1:
                identifier = records[0]["id"]
            elif len(records) < 1:
                raise Exception(
                    "No records found matching type and name - won't update"
                )
            else:
                raise Exception(
                    "Multiple records found matching type and name - won't update"
                )

        data = {}

        if name:
            data["name"] = name

        if rtype:
            data["type"] = rtype

        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        if content:
            if rtype == "TXT":
                content = '"' + content + '"'
            tmp = content
            content = []
            content.append(tmp)
            data["records"] = content

        self._put(f"/zones/{self.domain_id}/recordsets/{identifier}", data)
        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []

        tmp = content
        content = []
        content.append(tmp)

        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)
        for record_id in delete_record_id:
            self._delete(f"/zones/{self.domain_id}/recordsets/{record_id}")

        LOGGER.debug("delete_record: %s", True)
        return True

    # API requests
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "X-Auth-Token": f"{self._get_provider_option('auth_token')}",
        }
        if not url.startswith(self.api_endpoint):
            url = self.api_endpoint + url

        response = requests.request(
            action,
            url,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
        )
        response.raise_for_status()
        if action == "DELETE":
            return ""
        return response.json()



"""Module provider for Name.com"""

from __future__ import absolute_import

import logging
from argparse import ArgumentParser
from typing import List

from requests import HTTPError, Session
from requests.auth import HTTPBasicAuth

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

DUPLICATE_ERROR = {
    "message": "Invalid Argument",
    "details": "Parameter Value Error - Duplicate Record",
}


class NamecomLoader(
    object
):  # pylint: disable=useless-object-inheritance,too-few-public-methods
    """Loader that handles pagination for the Name.com provider."""

    def __init__(self, get, url, data_key, next_page=1):
        self.get = get
        self.url = url
        self.data_key = data_key
        self.next_page = next_page

    def __iter__(self):
        while self.next_page:
            response = self.get(self.url, {"page": self.next_page})
            for data in response[self.data_key]:
                yield data
            self.next_page = response.get("next_page")


class NamecomProvider(BaseProvider):
    """Provider implementation for Name.com."""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["name.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-username", help="specify a username")
        parser.add_argument("--auth-token", help="specify an API token")

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.api_endpoint = "https://api.name.com/v4"
        self.session = Session()

    def authenticate(self):
        self.session.auth = HTTPBasicAuth(
            username=self._get_provider_option("auth_username"),
            password=self._get_provider_option("auth_token"),
        )

        # checking domain existence
        domain_name = self.domain
        for domain in NamecomLoader(self._get, "/domains", "domains"):
            if domain["domainName"] == domain_name:
                self.domain_id = domain_name
                return

        raise AuthenticationError("{} domain does not exist".format(domain_name))

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype, name, content):
        data = {
            "type": rtype,
            "host": self._relative_name(name),
            "answer": content,
            "ttl": self._get_lexicon_option("ttl"),
        }

        if rtype in ("MX", "SRV"):
            # despite the documentation says a priority is
            # required for MX and SRV, it's actually optional
            priority = self._get_lexicon_option("priority")
            if priority:
                data["priority"] = priority

        url = "/domains/{}/records".format(self.domain)
        try:
            record_id = self._post(url, data)["id"]
        except HTTPError as error:
            response = error.response
            if response.status_code == 400 and response.json() == DUPLICATE_ERROR:
                LOGGER.warning("create_record: duplicate record has been skipped")
                return True
            raise

        LOGGER.debug("create_record: record %s has been created", record_id)

        return record_id

    def list_records(self, rtype=None, name=None, content=None):
        url = "/domains/{}/records".format(self.domain)
        records = []

        for raw in NamecomLoader(self._get, url, "records"):
            record = {
                "id": raw["id"],
                "type": raw["type"],
                "name": raw["fqdn"][:-1],
                "ttl": raw["ttl"],
                "content": raw["answer"],
            }
            records.append(record)

        LOGGER.debug("list_records: retrieved %s records", len(records))

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            name = self._full_name(name)
            records = [record for record in records if record["name"] == name]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: filtered %s records", len(records))

        return records

    def update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier:
            if not (rtype and name):
                raise ValueError("Record identifier or rtype+name must be specified")
            records = self.list_records(rtype, name)
            if not records:
                raise Exception("There is no record to update")

            if len(records) > 1:
                filtered_records = [
                    record for record in records if record["content"] == content
                ]
                if filtered_records:
                    records = filtered_records

                if len(records) > 1:
                    raise Exception(
                        "There are multiple records to update: {}".format(
                            ", ".join(record["id"] for record in records)
                        )
                    )

            record_id = records[0]["id"]
        else:
            record_id = identifier

        data = {"ttl": self._get_lexicon_option("ttl")}

        # even though the documentation says a type and an answer
        # are required, they are not required actually
        if rtype:
            data["type"] = rtype
        if name:
            data["host"] = self._relative_name(name)
        if content:
            data["answer"] = content

        url = "/domains/{}/records/{}".format(self.domain, record_id)
        record_id = self._put(url, data)["id"]
        logging.debug("update_record: record %s has been updated", record_id)

        return record_id

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if not identifier:
            if not (rtype and name):
                raise ValueError("Record identifier or rtype+name must be specified")
            records = self.list_records(rtype, name, content)
            if not records:
                LOGGER.warning("delete_record: there is no record to delete")
                return False
            record_ids = [record["id"] for record in records]
        else:
            record_ids = [
                identifier,
            ]

        for record_id in record_ids:
            url = "/domains/{}/records/{}".format(self.domain, record_id)
            self._delete(url)
            LOGGER.debug("delete_record: record %s has been deleted", record_id)

        return True

    def _get_raw_record(self, record_id):
        url = "/domains/{}/records/{}".format(self.domain, record_id)
        return self._get(url)

    def _request(self, action="GET", url="/", data=None, query_params=None):
        response = self.session.request(
            method=action, url=self.api_endpoint + url, json=data, params=query_params
        )
        response.raise_for_status()
        return response.json()


Provider = NamecomProvider



"""Module provider for Easyname DNS"""

import json
import logging
from argparse import ArgumentParser
from typing import List

from bs4 import BeautifulSoup, Tag
from requests import Response, Session

from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """
    easyname provider
    """

    URLS = {
        "login": "https://my.easyname.com/en/login",
        "login_endpoint": "https://my.easyname.com/en/authentication-api/login",
        "domain_list": "https://my.easyname.com/domains/",
        "dashboard": "https://my.easyname.com/en/dashboard",
        "overview": "https://my.easyname.com/hosting/view-user.php",
        "dns": "https://my.easyname.com/en/domain/dns/index/domain/{}/",
        "dns_create_entry": "https://my.easyname.com/en/domain/dns/create/domain/{}",
        "dns_update_entry": "https://my.easyname.com/en/domain/dns/edit/domain/{}/id/{}",
        "dns_delete_entry": "https://my.easyname.com/en/domain/dns/delete/domain/{}/id/{}",
        "dns_delete_entry_confirm": "https://my.easyname.com/en/domain/dns/delete/domain/{}/id/{}/confirm/1",
    }

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["easyname.eu"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.description = """A provider for Easyname DNS."""
        parser.add_argument(
            "--auth-username", help="Specify username used to authenticate"
        )
        parser.add_argument(
            "--auth-password",
            help="Specify password used to authenticate",
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.session = Session()
        self.domain_id = None
        self._records = None

    def _get_csrf_token(self):
        """Return the CSRF Token of easyname login form."""
        home_response = self.session.get(self.URLS["login"])
        self._log("Home", home_response)
        assert home_response.status_code == 200, "Could not load Easyname login page."

        csrf_token_field = home_response.cookies["CSRF-TOKEN"]

        assert csrf_token_field is not None, "Could not find login token."
        return csrf_token_field

    def _login(self, csrf_token):
        """Attempt to login session on easyname."""
        payload = {
            "emailAddress": self._get_provider_option("auth_username"),
            "password": self._get_provider_option("auth_password"),
        }
        headers = self.session.headers
        headers["X-CSRF-TOKEN"] = csrf_token

        login_response = self.session.post(
            self.URLS["login_endpoint"],
            data=json.dumps(payload),
            headers=headers,
            allow_redirects=True,
        )
        self._log("Login", login_response)
        json_response = json.loads(login_response.content)
        assert (
            login_response.status_code == 200
        ), "Could not login due to a network error."

        assert json_response["redirectUrl"] == self.URLS["dashboard"], (
            "Easyname login failed, bad EASYNAME_USER or EASYNAME_PASS.%s"
            % login_response.url
        )

    def authenticate(self):
        csrf_token = self._get_csrf_token()
        self._login(csrf_token)

        domain_text_element = self._get_domain_text_of_authoritative_zone()
        self.domain_id = self._get_domain_id(domain_text_element)
        LOGGER.debug("Easyname domain ID: %s", self.domain_id)

        return True

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype, name, content):
        return self._create_record_internal(rtype=rtype, name=name, content=content)

    def _create_record_internal(self, rtype, name, content, identifier=None):
        """
        Create a new DNS entry in the domain zone if it does not already exist.

        Args:
          rtype (str): The DNS type (e.g. A, TXT, MX, etc) of the new entry.
          name (str): The name of the new DNS entry, e.g the domain for which a
                      MX entry shall be valid.
          content (str): The content of the new DNS entry, e.g. the mail server
                         hostname for a MX entry.
          [identifier] (str): The easyname id of a DNS entry. Use to overwrite an
                    existing entry.

        Returns:
          bool: True if the record was created successfully, False otherwise.
        """
        name = self._relative_name(name) if name is not None else name
        LOGGER.debug("Creating record with name %s", name)
        if self._is_duplicate_record(rtype, name, content):
            return True

        data = self._get_post_data_to_create_dns_entry(rtype, name, content, identifier)
        if not identifier:
            LOGGER.debug("Create DNS data: %s", data)
            create_response = self.session.post(
                self.URLS["dns_create_entry"].format(self.domain_id), data=data
            )
            self._log("Create DNS entry", create_response)
        else:
            LOGGER.debug("Update DNS data: %s", data)
            update_response = self.session.post(
                self.URLS["dns_update_entry"].format(self.domain_id, identifier),
                data=data,
            )
            self._log("Update DNS entry", update_response)

        self._invalidate_records_cache()

        # Pull a list of records and check for ours
        was_success = len(self.list_records(rtype, name, content)) > 0
        if identifier != "" and was_success:
            msg = "Successfully updated record {}".format(name)
        elif was_success:
            msg = "Successfully added record {}".format(name)
        else:
            msg = "Failed to add/update record {}".format(name)
        LOGGER.info(msg, name)
        return was_success

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """
        Delete one or more DNS entries in the domain zone that match the given
        criteria.

        Args:
          [identifier] (str): An ID to match against DNS entry easyname IDs.
          [rtype] (str): A DNS rtype (e.g. A, TXT, MX, etc) to match against DNS
                      entry types.
          [name] (str): A name to match against DNS entry names.
          [content] (str): A content to match against a DNS entry contents.

        Returns:
          bool: True if the record(s) were deleted successfully, False
                otherwise.
        """
        record_ids = self._get_matching_dns_entry_ids(identifier, rtype, name, content)
        LOGGER.debug("Record IDs to delete: %s", record_ids)

        success = True
        for rec_id in record_ids:
            delete_response = self.session.get(
                self.URLS["dns_delete_entry"].format(self.domain_id, rec_id)
            )
            delete_response_confirm = self.session.post(
                self.URLS["dns_delete_entry_confirm"].format(self.domain_id, rec_id)
            )
            self._invalidate_records_cache()
            self._log("Delete DNS entry {rec_id}", delete_response)
            # success = success and delete_response_confirm.url == success_url
            success = "feedback-message--success" in delete_response_confirm.text

        return success

    def update_record(self, identifier, rtype=None, name=None, content=None):
        """
        Update a DNS entry identified by identifier or name in the domain zone.
        Any non given argument will leave the current value of the DNS entry.

        Args:
          identifier (str): The easyname id of the DNS entry to update.
          [rtype] (str): The DNS rtype (e.g. A, TXT, MX, etc) of the new entry.
          [name] (str): The name of the new DNS entry, e.g the domain for which
                        a MX entry shall be valid.
          [content] (str): The content of the new DNS entry, e.g. the mail
                           server hostname for a MX entry.

        Returns:
          bool: True if the record was updated successfully, False otherwise.

        Raises:
          AssertionError: When a request returns unexpected or unknown data.
        """
        if identifier is not None:
            identifier = int(identifier)
            records = self._list_records_internal(identifier=identifier)
        else:
            records = self._list_records_internal(name=name, rtype=rtype)
        LOGGER.debug("Records to update (%d): %s", len(records), records)
        assert records, "No record found to update"
        success = True

        for record in records:
            name = name if name is not None else record["name"]
            rtype = rtype if rtype is not None else record["type"]
            content = content if content is not None else record["content"]
            success = success and self._create_record_internal(
                rtype, name, content, record["id"]
            )
        return success

    def list_records(self, rtype=None, name=None, content=None):
        return self._list_records_internal(rtype=rtype, name=name, content=content)

    def _list_records_internal(
        self, rtype=None, name=None, content=None, identifier=None
    ):
        """
        Filter and list DNS entries of domain zone on Easyname.
        Easyname shows each entry in a HTML table row and each attribute on a
        table column.

        Args:
          [rtype] (str): Filter by DNS rtype (e.g. A, TXT, MX, etc)
          [name] (str): Filter by the name of the DNS entry, e.g the domain for
                      which a MX entry shall be valid.
          [content] (str): Filter by the content of the DNS entry, e.g. the
                           mail server hostname for a MX entry.
          [identifier] (str): Filter by the easyname id of the DNS entry.

        Returns:
          list: A list of DNS entries. A DNS entry is an object with DNS
                attribute names as keys (e.g. name, content, priority, etc)
                and additionally an id.

        Raises:
          AssertionError: When a request returns unexpected or unknown data.
        """
        name = self._full_name(name) if name is not None else name
        if self._records is None:
            records = []
            # skip the first record which contains the table header
            rows = self._get_dns_entry_trs()[1:]

            for row in rows:
                self._log("DNS list entry", row)
                try:
                    rec = {}
                    columns = row.find_all("td")
                    remove_span_from_name = columns[0].text.replace(
                        '<span style="color: var(--color-grey-500);">', ""
                    )
                    columns[0].replace_with(remove_span_from_name)
                    rec["name"] = (remove_span_from_name or "").strip()
                    rec["type"] = (columns[1].contents[1].text or "").strip()
                    rec["content"] = (columns[2].contents[1].string or "").strip()
                    rec["priority"] = (columns[3].contents[1].string or "").strip()
                    rec["ttl"] = (columns[4].contents[1].string or "").strip()
                    rec["id"] = ""
                    for a in columns[5].find_all(
                        "a", class_="button button--transparent"
                    ):
                        rec["id"] = int(a["href"].rsplit("/", 1)[-1])
                    if rec["priority"]:
                        rec["priority"] = int(rec["priority"])
                    if rec["ttl"]:
                        rec["ttl"] = int(rec["ttl"])
                except Exception as error:
                    errmsg = "Cannot parse DNS entry ({}).".format(error)
                    LOGGER.warning(errmsg)
                    raise AssertionError(errmsg)
                records.append(rec)
            self._records = records
        records = self._filter_records(self._records, rtype, name, content, identifier)
        LOGGER.debug("Final records (%d): %s", len(records), records)
        return records

    def _request(self, action="GET", url="/", data=None, query_params=None):
        pass

    def _invalidate_records_cache(self):
        """
        Invalidate DNS entries cache such that list_records will do a new
        request to retrieve DNS entries.
        """
        self._records = None

    def _get_post_data_to_create_dns_entry(self, rtype, name, content, identifier=None):
        """
        Build and return the post date that is needed to create a DNS entry.
        """
        is_update = identifier is not None
        record = None
        if is_update:
            records = self._list_records_internal(identifier=identifier)
            assert len(records) == 1, "ID is not unique or does not exist"
            record = records[0]
            LOGGER.debug("Create post data to update record: %s", record)

        data = {
            "id": str(identifier) if is_update else "",
            "name": name,
            "type": rtype,
            "content": content,
            "priority": str(record["priority"]) if is_update else "10",
            "ttl": str(record["ttl"]) if is_update else "360",
        }
        ttl = self._get_lexicon_option("ttl")
        if ttl and ttl > 360:
            data["ttl"] = str(ttl)

        prio = self._get_lexicon_option("priority")
        if prio and prio > 0:
            data["priority"] = str(prio)

        return data

    def _is_duplicate_record(self, rtype, name, content):
        """Check if DNS entry already exists."""
        records = self.list_records(rtype, name, content)
        is_duplicate = len(records) >= 1
        if is_duplicate:
            LOGGER.info("Duplicate record %s %s %s, NOOP", rtype, name, content)
        return is_duplicate

    def _get_matching_dns_entry_ids(
        self, identifier=None, rtype=None, name=None, content=None
    ):
        """Return a list of DNS entries that match the given criteria."""
        record_ids = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            record_ids = [record["id"] for record in records]
        else:
            record_ids.append(identifier)
        return record_ids

    def _get_dns_entry_trs(self):
        """
        Return the TR elements holding the DNS entries.
        """
        dns_list_response = self.session.get(self.URLS["dns"].format(self.domain_id))
        self._log("DNS list", dns_list_response)
        assert dns_list_response.status_code == 200, "Could not load DNS entries."

        html = BeautifulSoup(dns_list_response.content, "html.parser")
        self._log("DNS list", html)
        dns_table = html.find("table")
        assert dns_table is not None, "Could not find DNS entry table"

        def _is_zone_tr(elm):
            return elm.name.lower() == "tr" and (elm.has_attr("class"))

        rows = dns_table.find_all(_is_zone_tr)
        assert rows is not None and rows, "Could not find any DNS entries"
        return rows

    def _filter_records(
        self, records, rtype=None, name=None, content=None, identifier=None
    ):
        """
        Filter dns entries based on type, name or content.
        """
        if not records:
            return []
        if identifier is not None:
            LOGGER.debug("Filtering %d records by id: %s", len(records), identifier)
            records = [record for record in records if record["id"] == identifier]
        if rtype is not None:
            LOGGER.debug("Filtering %d records by type: %s", len(records), rtype)
            records = [record for record in records if record["type"] == rtype]
        if name is not None:
            LOGGER.debug("Filtering %d records by name: %s", len(records), name)
            if name.endswith("."):
                name = name[:-1]
            if rtype.lower() != "txt":
                name = name.split(".")[0]
                records = [
                    record for record in records if name == record["name"].split(".")[0]
                ]
            else:
                records = [record for record in records if name == record["name"]]
        if content is not None:
            LOGGER.debug(
                "Filtering %d records by content: %s", len(records), content.lower()
            )
            records = [
                record
                for record in records
                if record["content"].lower() == content.lower()
            ]
        return records

    def _get_domain_text_of_authoritative_zone(self):
        """Get the authoritative name zone."""
        # We are logged in, so get the domain list
        zones_response = self.session.get(self.URLS["domain_list"])
        self._log("Zone", zones_response)
        assert (
            zones_response.status_code == 200
        ), "Could not retrieve domain list due to a network error."

        html = BeautifulSoup(zones_response.content, "html.parser")
        self._log("Zone", html)
        domain_table = html.find("table")
        assert domain_table is not None, "Could not find domain table"

        # (Sub)domains can either be managed in their own zones or by the
        # zones of their parent (sub)domains. Iterate over all subdomains
        # (starting with the deepest one) and see if there is an own zone
        # for it.
        domain = self.domain or ""
        domain_text = None
        subdomains = domain.split(".")
        while True:
            domain = ".".join(subdomains)
            LOGGER.debug("Check if %s has own zone", domain)
            domain_text = domain_table.find(string=domain)
            if domain_text is not None or len(subdomains) < 3:
                break
            subdomains.pop(0)

        # Update domain to equal the zone's domain. This is important if we are
        # handling a subdomain that has no zone of itself. If we do not do
        # this, self._relative_name will strip also a part of the subdomain
        # away.
        self.domain = domain
        assert domain_text is not None, "The domain does not exist on Easyname."
        return domain_text

    def _get_domain_id(self, domain_text_element):
        """Return the easyname id of the domain."""
        try:
            # Hierarchy: TR > TD > SPAN > Domain Text
            tr_anchor = domain_text_element.parent.parent.parent
            td_anchor = tr_anchor.find("td", {"class": "entity__field taright"})
            link = td_anchor.find("a")["href"]
            domain_id = link.rsplit("/", 1)[-1]
            return domain_id
        except Exception as error:
            errmsg = (
                "Cannot get the domain id even though the domain seems "
                "to exist (%s).",
                error,
            )
            LOGGER.warning(errmsg)
            raise AssertionError(errmsg)

    def _log(self, name, element):
        """
        Log Response and Tag elements. Do nothing if elements is none of them.
        """
        if isinstance(element, Response):
            LOGGER.debug(
                "%s response: URL=%s Code=%s", name, element.url, element.status_code
            )
        elif isinstance(element, (BeautifulSoup, Tag)):
            LOGGER.debug("%s HTML:\n%s", name, element)



import json
import logging
from argparse import ArgumentParser
from typing import List

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

# tencentcloud-sdk-python is an optional dependency of lexicon; do not throw an ImportError if
# the dependency is unmet.
try:
    from tencentcloud.common import credential  # type: ignore
    from tencentcloud.dnspod.v20210323 import dnspod_client, models  # type: ignore
except ImportError:
    pass

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for QCloud"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["dnspod.tencentcloudapi.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--secret_id", help="specify secret_id for authentication")
        parser.add_argument(
            "--secret_key", help="specify secret_key for authentication"
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        secret_id = self._get_provider_option("secret_id")
        secret_key = self._get_provider_option("secret_key")
        if not secret_id:
            raise Exception("Error, application secret_id is not defined")

        if not secret_key:
            raise Exception("Error, application secret_key is not defined")

        self.cred = credential.Credential(secret_id, secret_key)
        self.client = dnspod_client.DnspodClient(self.cred, "")
        # add support for other api entrypoints

    def authenticate(self):
        describe_domain_req = models.DescribeDomainListRequest()
        params = {
            "Keyword": self.domain,
        }
        describe_domain_req.from_json_string(json.dumps(params))
        describe_domain_resp = self.client.DescribeDomainList(describe_domain_req)
        if len(describe_domain_resp.DomainList) == 0:
            raise AuthenticationError(
                f"Authenticate failed, domain {self.domain} not found"
            )
        self.domain_id = describe_domain_resp.DomainList[0].DomainId

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype, name, content):
        create_record_req = models.CreateRecordRequest()
        params = {
            "Domain": self.domain,
            "SubDomain": self._relative_name(name),
            "RecordType": rtype,
            "RecordLine": "\u9ed8\u8ba4",
            "Value": content,
        }
        if self._get_lexicon_option("ttl"):
            params["TTL"] = self._get_lexicon_option("ttl")
        create_record_req.from_json_string(json.dumps(params))
        LOGGER.debug("create_record req: %s", create_record_req.to_json_string())
        create_record_resp = self.client.CreateRecord(create_record_req)
        LOGGER.debug("create_record resp: %s", create_record_resp.to_json_string())
        return True

    def list_records(self, rtype=None, name=None, content=None):
        list_record_req = models.DescribeRecordListRequest()
        params = {
            "Domain": self.domain,
        }
        if rtype is not None:
            params["Type"] = rtype
        if name is not None:
            params["Subdomain"] = self._relative_name(name)
        # qcloud dns api not support filter by content, implement it by client filter

        list_record_req.from_json_string(json.dumps(params))
        LOGGER.debug("list_records req: %s", list_record_req.to_json_string())
        list_record_resp = self.client.DescribeRecordList(list_record_req)
        LOGGER.debug("list_records resp: %s", list_record_resp.to_json_string())

        records = []
        for record in list_record_resp.RecordList:
            if content is not None and record.Value != content:
                continue
            records.append(
                {
                    "type": record.Type,
                    "name": self._full_name(record.Name),
                    "ttl": record.TTL,
                    "content": record.Value,
                    "id": record.RecordId,
                }
            )
        return records

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        batch_delete_record_req = models.DeleteRecordBatchRequest()
        params = {}
        if not identifier:
            records = self.list_records(rtype, name, content)
            if len(records) == 0:
                return True
            params["RecordIdList"] = [record["id"] for record in records]
        else:
            params["RecordIdList"] = [identifier]
        batch_delete_record_req.from_json_string(json.dumps(params))

        LOGGER.debug("delete_record req: %s", batch_delete_record_req.to_json_string())
        batch_delete_record_resp = self.client.DeleteRecordBatch(
            batch_delete_record_req
        )
        LOGGER.debug(
            "delete_record resp: %s", batch_delete_record_resp.to_json_string()
        )
        return True

    def update_record(self, identifier, rtype=None, name=None, content=None):
        modify_record_req = models.ModifyRecordRequest()
        params = {
            "Domain": self.domain,
            "SubDomain": self._relative_name(name),
            "RecordType": rtype,
            "RecordLine": "\u9ed8\u8ba4",
            "Value": content,
            "RecordId": identifier,
        }
        if self._get_lexicon_option("ttl"):
            params["TTL"] = self._get_lexicon_option("ttl")

        modify_record_req.from_json_string(json.dumps(params))

        LOGGER.debug("update_record req: %s", modify_record_req.to_json_string())
        modify_record_resp = self.client.ModifyRecord(modify_record_req)
        LOGGER.debug("update_record resp: %s", modify_record_resp.to_json_string())

        return True



"""Module provider for Aurora"""

import base64
import datetime
import hashlib
import hmac
import json
import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider for Aurora"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["auroradns.eu"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-api-key", help="specify API key for authentication")
        parser.add_argument(
            "--auth-secret-key", help="specify the secret key for authentication"
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.auroradns.eu"

    def authenticate(self):
        payload = self._get("/zones")

        for item in payload:
            if item["name"] == self.domain:
                self.domain_id = item["id"]
                break
        else:
            raise AuthenticationError("No domain found")

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, rtype, name, content):
        data = {"type": rtype, "name": self._relative_name(name), "content": content}
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")
        payload = self._post(f"/zones/{self.domain_id}/records", data)

        LOGGER.debug("create_record: %s", payload)
        return payload

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        payload = self._get(f"/zones/{self.domain_id}/records")

        # Apply filtering first.
        processed_records = payload
        if rtype:
            processed_records = [
                record for record in processed_records if record["type"] == rtype
            ]
        if name:
            processed_records = [
                record
                for record in processed_records
                if record["name"] == self._relative_name(name)
            ]
        if content:
            processed_records = [
                record
                for record in processed_records
                if record["content"].lower() == content.lower()
            ]

        # Format the records.
        records = []
        for record in processed_records:
            processed_record = {
                "type": record["type"],
                "name": self._full_name(record["name"]),
                "ttl": record["ttl"],
                "content": record["content"],
                "id": record["id"],
            }
            records.append(processed_record)

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        # Try to find record if no identifier was specified
        if not identifier:
            identifier = self._find_record_identifier(rtype, name, None)

        data = {}
        if rtype:
            data["type"] = rtype
        if name:
            data["name"] = self._relative_name(name)
        if content:
            data["content"] = content
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        payload = self._put(f"/zones/{self.domain_id}/records/{identifier}", data)

        LOGGER.debug("update_record: %s", payload)
        return payload

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        # Try to find record if no identifier was specified
        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete(f"/zones/{self.domain_id}/records/{record_id}")

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        time = datetime.datetime.now(datetime.timezone.utc)
        timestamp = time.strftime("%Y%m%dT%H%M%SZ")
        authorization_header = self._generate_auth_header(action, url, timestamp)

        request = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers={
                "X-AuroraDNS-Date": timestamp,
                "Authorization": authorization_header,
                "Content-Type": "application/json",
            },
        )

        # If the response is a HTTP 409 statusCode, the record already exists: return true.
        if request.status_code == 409:
            return True

        # If the request fails for any other reason, throw an error.
        request.raise_for_status()

        # Try to parse the json, if it not exists, return true.
        try:
            return request.json()
        except BaseException:
            return True

    def _generate_auth_header(self, action, url, timestamp):
        secret_key = self._get_provider_option("auth_secret_key")
        api_key = self._get_provider_option("auth_api_key")
        sig = action + url + timestamp

        signature = base64.b64encode(
            hmac.new(
                secret_key.encode("utf-8"),
                sig.encode("utf-8"),
                digestmod=hashlib.sha256,
            ).digest()
        )

        auth = api_key + ":" + signature.decode("utf-8")
        auth_b64 = base64.b64encode(auth.encode("utf-8"))
        return f"AuroraDNSv1 {auth_b64.decode('utf-8')}"

    def _find_record_identifier(self, rtype, name, content):
        records = self.list_records(rtype, name, content)
        LOGGER.debug("records: %s", records)
        if len(records) == 1:
            return records[0]["id"]
        raise Exception(
            "Record identifier could not be found. Try to provide an identifier"
        )



"""Module provider for INWX"""

import logging
from argparse import ArgumentParser
from typing import List

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

try:
    import xmlrpclib  # type: ignore
except ImportError:
    import xmlrpc.client as xmlrpclib

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """
    INWX offers a free testing system on https://ote.inwx.com
    see https://www.inwx.de/en/offer/api for details about ote and the api
    """

    @staticmethod
    def get_nameservers() -> List[str]:
        return [
            "ns.inwx.de",
            "ns2.inwx.de",
            "ns3.inwx.eu",
            "ns4.inwx.com",
            "ns5.inwx.net",
            "ns.domrobot.com",
            "ns.domrobot.net",
            "ns.domrobot.org",
            "ns.domrobot.info",
            "ns.domrobot.biz",
        ]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-username", help="specify username for authentication"
        )
        parser.add_argument(
            "--auth-password", help="specify password for authentication"
        )

    def __init__(self, config):
        """
        :param config: command line options
        """
        super(Provider, self).__init__(config)

        self._auth = {
            "user": self._get_provider_option("auth_username"),
            "pass": self._get_provider_option("auth_password"),
        }
        self._domain = self.domain.lower()
        self.domain_id = None

        endpoint = (
            self._get_provider_option("endpoint") or "https://api.domrobot.com/xmlrpc/"
        )

        self._api = xmlrpclib.ServerProxy(endpoint, allow_none=True)

    def _validate_response(self, response, message, exclude_code=None):
        """
        validate an api server response

        :param dict response: server response to check
        :param str message: error message to raise
        :param int exclude_code: error codes to exclude from errorhandling
        :return:
        ":raises Exception: on error
        """
        if "code" in response and response["code"] >= 2000:
            if exclude_code is not None and response["code"] == exclude_code:
                return

            raise Exception(f"{message}: {response['msg']} ({response['code']})")

    # Make any request to validate credentials
    def authenticate(self):
        """
        run any request against the API just to make sure the credentials
        are valid

        :return bool: success status
        :raises Exception: on error
        """
        opts = {"domain": self._domain}
        opts.update(self._auth)
        response = self._api.nameserver.info(opts)
        try:
            self._validate_response(response=response, message="Failed to authenticate")
        except Exception as e:
            raise AuthenticationError(str(e))

        # set to fake id to pass tests, inwx doesn't work on domain id but
        # uses domain names for identification
        self.domain_id = 1

        return True

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype, name, content):
        """
        create a record
        does nothing if the record already exists

        :param str rtype: type of record
        :param str name: name of record
        :param mixed content: value of record
        :return bool: success status
        :raises Exception: on error
        """
        opts = {
            "domain": self._domain,
            "type": rtype.upper(),
            "name": self._full_name(name),
            "content": content,
        }
        if self._get_lexicon_option("ttl"):
            opts["ttl"] = self._get_lexicon_option("ttl")
        opts.update(self._auth)

        response = self._api.nameserver.createRecord(opts)
        self._validate_response(
            response=response, message="Failed to create record", exclude_code=2302
        )

        return True

    def list_records(self, rtype=None, name=None, content=None):
        """
        list all records

        :param str rtype: type of record
        :param str name: name of record
        :param mixed content: value of record
        :return list: list of found records
        :raises Exception: on error
        """
        opts = {"domain": self._domain}
        if rtype is not None:
            opts["type"] = rtype.upper()
        if name is not None:
            opts["name"] = self._full_name(name)
        if content is not None:
            opts["content"] = content
        opts.update(self._auth)

        response = self._api.nameserver.info(opts)
        self._validate_response(response=response, message="Failed to get records")

        records = []
        if "record" in response["resData"]:
            for record in response["resData"]["record"]:
                processed_record = {
                    "type": record["type"],
                    "name": record["name"],
                    "ttl": record["ttl"],
                    "content": record["content"],
                    "id": record["id"],
                }
                records.append(processed_record)

        return records

    def update_record(self, identifier, rtype=None, name=None, content=None):
        """
        update a record

        :param int identifier: identifier of record to update
        :param str rtype: type of record
        :param str name: name of record
        :param mixed content: value of record
        :return bool: success status
        :raises Exception: on error
        """
        record_ids = []
        if not identifier:
            records = self.list_records(rtype, name)
            record_ids = [record["id"] for record in records]
        else:
            record_ids.append(identifier)

        for an_identifier in record_ids:
            opts = {"id": an_identifier}
            if rtype is not None:
                opts["type"] = rtype.upper()
            if name is not None:
                opts["name"] = self._full_name(name)
            if content is not None:
                opts["content"] = content
            opts.update(self._auth)

            response = self._api.nameserver.updateRecord(opts)
            self._validate_response(
                response=response, message="Failed to update record", exclude_code=2302
            )

        return True

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        """
        delete a record
        filter selection to delete by identifier or rtype/name/content

        :param int identifier: identifier of record to update
        :param str rtype: rtype of record
        :param str name: name of record
        :param mixed content: value of record
        :return bool: success status
        :raises Exception: on error
        """
        record_ids = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            record_ids = [record["id"] for record in records]
        else:
            record_ids.append(identifier)

        for record_id in record_ids:
            opts = {"id": record_id}
            opts.update(self._auth)
            response = self._api.nameserver.deleteRecord(opts)
            self._validate_response(
                response=response, message="Failed to update record"
            )

        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # Helper _request is not used for INWX provider.
        pass



"""Module provider for deSEC"""

import hashlib
import logging
import re
from argparse import ArgumentParser
from typing import Optional, Tuple

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import HTTPError
from urllib3.util.retry import Retry

from lexicon._private.discovery import lexicon_version
from lexicon.config import ConfigResolver
from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

# Type aliases
RecordType = Tuple[dict, int]
RecordList = dict[str, RecordType]
StrDict = dict[str, str]
StrDictList = list[StrDict]
OptStr = Optional[str]
OptStrDict = Optional[StrDict]
SanitizedResponseType = Tuple[str, str, dict]


class Provider(BaseProvider):
    """Provider class for deSEC"""

    @staticmethod
    def get_nameservers() -> list[str]:
        return ["ns1.desec.io", "ns2.desec.org"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="specify api token for authentication")
        parser.add_argument(
            "--auth-username", help="specify email address for authentication"
        )
        parser.add_argument(
            "--auth-password", help="specify password for authentication"
        )

    def __init__(self, config: ConfigResolver):
        super(Provider, self).__init__(config)
        self.api_endpoint = "https://desec.io/api/v1"
        self._lexicon_version = lexicon_version()
        self._token = self._get_provider_option("auth_token")
        self._priority = self._get_lexicon_option("priority")
        if self._priority and not self._priority.isnumeric():
            raise ValueError(f"Priority argument '{self._priority}' is not numeric.")

        # RegEx patterns, priority optional
        self._re = {
            "MX": re.compile(r"((?P<priority>\d+)\s+)?(?P<target>.+)"),
            "SRV": re.compile(
                r"((?P<priority>\d+)\s+)?(?P<weight>\d+)\s+(?P<port>\d+)\s+(?P<target>.+)"
            ),
        }

        # deSEC enforces rate limits, which are hit by rapid successive requests,
        # like by the automated tests via pytest. The API responses with status 429
        # and a retry-after header, which we want to use for retries.

        # https://desec.readthedocs.io/en/latest/rate-limits.html
        # dns_api_per_domain_expensive: 2/s - 15/min - 100/h - 300/day
        self._session = requests.Session()
        self._session.mount(
            "https://",
            HTTPAdapter(
                max_retries=Retry(
                    allowed_methods=None,  # Allow all methods
                    respect_retry_after_header=True,
                )
            ),
        )

    def authenticate(self):
        # Handle authentication
        username = self._get_provider_option("auth_username")
        password = self._get_provider_option("auth_password")
        if self._token and (username or password):
            raise AuthenticationError("Multiple authentication mechanisms specified.")
        if username and password and not self._token:
            self._login(username, password)
        if not self._token:
            raise AuthenticationError("No valid authentication mechanism specified.")

        # Check if domain exists
        try:
            self._get("..")
            self.domain_id = self.domain
            LOGGER.debug(f"authenticate: domain '{self.domain}' found.")
        except HTTPError as err:
            raise AuthenticationError(
                f"Domain '{self.domain}' not found ({err.response.status_code})."
            )

    def cleanup(self) -> None:
        pass

    def list_records(
        self, rtype: OptStr = None, name: OptStr = None, content: OptStr = None
    ) -> list[dict]:
        rec_sets = self._fetch_record_sets(rtype, name, content)
        records = [
            {
                "type": rec_set["type"],
                "ttl": rec_set["ttl"],
                "name": self._format_name(rec_set),
                "id": identifier,
                "content": sanitized,
            }
            | options
            for identifier, (rec_set, index) in rec_sets.items()
            for sanitized, dirty, options in [
                self._sanitize_response_content(rec_set, index)
            ]
        ]
        LOGGER.debug("list_records: %s", records)
        return records

    # Create record. If record already exists with the same content, do nothing
    def create_record(self, rtype: str, name: str, content: str) -> bool:
        (desec_rec, index) = self._get_record_set(rtype, name)
        desec_content = self._sanitize_request_content(content, rtype)
        subname = self._relative_name(name) if name else ""
        if not desec_rec or index == -1:
            # Create new record set, as it doesn't exist yet
            LOGGER.debug("_create_record: Creating new record set")
            self._post(
                "",
                {
                    "records": [desec_content],
                    "type": rtype,
                    "subname": subname,
                },
            )
            return True

        # Check if record is in record set
        desec_records: list[str] = desec_rec["records"]
        if content in desec_records or desec_content in desec_records:
            LOGGER.debug("_create_record: The record already exists. Ignore.")
            return True

        # Patch existing record set
        desec_records.append(desec_content)
        self._patch(f"{subname}/{rtype}", desec_rec)
        return True

    # Update a record.
    def update_record(
        self,
        identifier: OptStr = None,
        rtype: OptStr = None,
        name: OptStr = None,
        content: OptStr = None,
    ) -> bool:
        if identifier and name:
            # We can only filter for record type, as the subname could differ
            rec_sets = self._fetch_record_sets(rtype)
            if identifier not in rec_sets:
                LOGGER.warning(
                    f"update_record: No match for identifier '{identifier}'. Abort."
                )
                return False
            (desec_rec, index) = rec_sets[identifier]
        else:
            # We can't filter for content, as it likely changed
            (desec_rec, index) = self._get_record_set(rtype, name, None, identifier)

        if not desec_rec or index == -1:
            LOGGER.warning("update_record: No matching record found. Abort.")
            return False

        desec_records: list[str] = desec_rec["records"]
        rtype = rtype or desec_rec["type"]
        content = content or desec_records[index]

        # The subname is immutable and can't be modified.
        # We need to delete the old record and create a new one.
        old_subname = desec_rec["subname"]
        new_subname = self._relative_name(name) if name else ""
        if old_subname != new_subname:
            LOGGER.debug(
                f"update_record: new subname '{new_subname}' differs from old '{old_subname}'. Delete and recreate record."
            )
            return self._delete_record(
                desec_rec, index, rtype, old_subname
            ) and self.create_record(rtype, new_subname, content)

        # Patch the content
        new_content = self._sanitize_request_content(content or "", rtype or "")
        desec_records[index] = new_content
        self._patch(f"{new_subname or '@'}/{rtype}", desec_rec)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(
        self,
        identifier: OptStr = None,
        rtype: OptStr = None,
        name: OptStr = None,
        content: OptStr = None,
    ) -> bool:
        # Get first item
        (desec_rec, index) = self._get_record_set(rtype, name, content, identifier)
        if not desec_rec or index == -1:
            LOGGER.debug("delete_record: Record not found. Ignore.")
            return True
        subname = desec_rec["subname"]
        rtype = rtype or desec_rec["type"]
        if rtype and name and not (content or identifier):
            LOGGER.debug(
                f"delete_record: remove whole '{rtype}' record set '{subname}'"
            )
            self._delete(f"{subname}/{rtype}/")
            return True
        return self._delete_record(desec_rec, index, rtype, subname)

    def _delete_record(
        self, desec_rec: dict, index: int, rtype: str, subname: str
    ) -> bool:
        # Delete specific record
        desec_records: list[str] = desec_rec["records"]
        removed = desec_records.pop(index)
        self._patch(f"{subname}/{rtype}", desec_rec)
        LOGGER.debug(
            f"_delete_record: removed '{rtype}' item from '{subname}' record set, with content '{removed}'"
        )
        return True

    # Helpers

    def _request(
        self,
        action: str = "GET",
        url: str = "/",
        data: OptStrDict = None,
        query_params: OptStrDict = None,
    ):
        # TTL is required for all deSEC record sets
        if data:
            if ttl := self._get_lexicon_option("ttl"):
                data["ttl"] = ttl
            if not data["ttl"]:
                data["ttl"] = "3600"

        response = self._session.request(
            action,
            f"{self.api_endpoint}/domains/{self.domain}/rrsets/{url}/",
            params=query_params,
            json=data,
            headers={
                "Accept": "application/json",
                "Content-Type": "application/json",
                "User-Agent": f"lexicon/{self._lexicon_version} desec",
                "Authorization": f"Token {self._token}",
            },
        )

        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        # DELETE responses without a body
        if response.status_code == 204:
            return {}
        return response.json()

    def _login(self, username: str, password: str) -> None:
        LOGGER.debug("_login: logging in with username / password")
        auth_res = requests.post(
            self.api_endpoint + "/auth/login/",
            None,
            {"email": username, "password": password},
        )
        auth_res.raise_for_status()

        json_res = auth_res.json()
        if json_res.get("mfa"):
            raise AuthenticationError("Login with enabled MFA/2FA is not supported.")

        self._token = json_res.get("token")
        if not self._token:
            raise AuthenticationError("Login successful, but no token was acquired.")

    def _fetch_record_sets(
        self, rtype: OptStr = None, name: OptStr = None, content: OptStr = None
    ) -> RecordList:
        desec_content = self._sanitize_request_content(content or "", rtype or "")
        response = self._get(
            "",
            {
                "type": rtype or None,
                "subname": self._relative_name(name) if name else None,
            },
        )

        # Generates a dict with the identifier as key
        # and a tuple of the record set and the index of related record
        # {id: (rec_set, index)}
        id_sets = {
            self._identifier(rec_set, rec): (rec_set, i)
            for rec_set in response
            for i, rec in enumerate(rec_set["records"])
            if (not content or content == rec)
            or (not desec_content or desec_content == rec)
        }
        LOGGER.debug("_fetch_record_sets: %s", id_sets)
        return id_sets

    def _get_record_set(
        self,
        rtype: OptStr = None,
        name: OptStr = None,
        content: OptStr = None,
        identifier: OptStr = None,
    ) -> RecordType:
        rec_sets = self._fetch_record_sets(rtype, name, content)
        if not len(rec_sets):
            # No record set found
            LOGGER.debug("_get_record_set: No match.")
            return ({}, -1)

        rec_set, *_ = rec_sets.values()
        if identifier and identifier in rec_sets:
            # Return specified record set
            return rec_sets[identifier]

        # Return first result
        return rec_set

    # Override, handle apex
    def _relative_name(self, record_name: str) -> str:
        subname = super()._relative_name(record_name or "@")
        return subname if subname != "@" else ""

    # Override, allow foreign domains
    def _fqdn_name(self, record_name: str) -> str:
        return (
            record_name
            if record_name.endswith(".")
            else super()._fqdn_name(record_name)
        )

    @staticmethod
    def _format_name(match: StrDict) -> str:
        sub = match["subname"]
        return f"{'@.' if not sub else ''}{match['name']}".strip(".")

    @staticmethod
    def _identifier(match: StrDict, record: str) -> str:
        sha256 = hashlib.sha256()
        sha256.update(f"{match['name']} => {match['type']} => '{record}'".encode())
        return sha256.hexdigest()[0:7]

    def _sanitize_request_content(self, content: str, rtype: str) -> str:
        if rtype == "TXT":
            content = content.strip('"')
            return f'"{content}"' if content else ""
        if rtype == "CNAME":
            return self._fqdn_name(content)
        if rtype in ("MX", "SRV"):
            # The priority is only relevant for MX and SRV types.
            # deSEC does not support this property, it is part of the record's content.
            parsed = self._parse_priority_record(content, rtype)
            priority = parsed.get("priority") or str(self._priority)
            parsed["priority"] = priority  # Ensure fallback for join operation
            if not priority:
                raise ValueError("Priority value is not defined.")
            if self._priority and self._priority != priority:
                raise ValueError(
                    f"The priority was specified as an argument ({self._priority}) "
                    f"and in the content ({priority}), but it doesn't match."
                )
            return " ".join(parsed.values())
        return content

    def _sanitize_response_content(
        self, rec_set: dict, index: int
    ) -> SanitizedResponseType:
        rtype = rec_set["type"]
        content = rec_set["records"][index]
        if rtype in ("MX", "SRV"):
            parsed = self._parse_priority_record(content, rtype)
            if not (priority := parsed.get("priority")) or not priority.isnumeric():
                raise Exception("Priority value is not present in content.")
            # Convert numeric options to int, see `technical_workbook.rst`
            options: dict = {
                k: (int(v) if v.isnumeric() else v) for k, v in parsed.items()
            }
            return " ".join(parsed.values()), content, {rtype.lower(): options}
        return content.strip('"'), content, {}

    def _parse_priority_record(self, content: str, rtype: str) -> StrDict:
        if not (match := self._re[rtype].match(content)):
            raise Exception(f"Content '{content}' is not valid for type '{rtype}'.")

        groups = match.groupdict()
        groups["target"] = self._fqdn_name(groups["target"])
        return groups
